import itertools
from typing import Any, Callable

import pytest

from tests.utils import decimal_to_int
from vyper.compiler import compile_code
from vyper.exceptions import (
    ArgumentException,
    ArrayIndexException,
    ImmutableViolation,
    OverflowException,
    StackTooDeep,
    StateAccessViolation,
    TypeMismatch,
)


def test_list_tester_code(get_contract):
    list_tester_code = """
z: DynArray[int128, 3]
z2: DynArray[DynArray[int128, 2], 2]
z3: DynArray[int128, 2]

@external
def foo(x: DynArray[int128, 3]) -> int128:
    return x[0] + x[1] + x[2]

@external
def goo(x: DynArray[DynArray[int128, 2], 2]) -> int128:
    return x[0][0] + x[0][1] + x[1][0] * 10 + x[1][1] * 10

@external
def hoo(x: DynArray[int128, 3]) -> int128:
    y: DynArray[int128, 3] = x
    return y[0] + x[1] + y[2]

@external
def joo(x: DynArray[DynArray[int128, 2], 2]) -> int128:
    y: DynArray[DynArray[int128, 2], 2] = x
    y2: DynArray[int128, 2] = x[1]
    return y[0][0] + y[0][1] + y2[0] * 10 + y2[1] * 10

@external
def koo(x: DynArray[int128, 3]) -> int128:
    self.z = x
    return self.z[0] + x[1] + self.z[2]

@external
def loo(x: DynArray[DynArray[int128, 2], 2]) -> int128:
    self.z2 = x
    self.z3 = x[1]
    return self.z2[0][0] + self.z2[0][1] + self.z3[0] * 10 + self.z3[1] * 10
    """

    c = get_contract(list_tester_code)
    assert c.foo([3, 4, 5]) == 12
    assert c.goo([[1, 2], [3, 4]]) == 73
    assert c.hoo([3, 4, 5]) == 12
    assert c.joo([[1, 2], [3, 4]]) == 73
    assert c.koo([3, 4, 5]) == 12
    assert c.loo([[1, 2], [3, 4]]) == 73
    print("Passed list tests")


def test_string_list(get_contract):
    code = """
@external
def foo1(x: DynArray[String[32], 2]) -> DynArray[String[32], 2]:
    return x

@external
def foo2(x: DynArray[DynArray[String[32], 2], 2]) -> DynArray[DynArray[String[32], 2], 2]:
    return x

@external
def foo3(x: DynArray[DynArray[String[32], 2], 2]) -> DynArray[String[32], 2]:
    return x[0]

@external
def foo4(x: DynArray[DynArray[String[32], 2], 2]) -> String[32]:
    return x[0][0]

@external
def foo5() -> DynArray[String[32], 2]:
    ret: DynArray[String[32], 2] = ["hello"]
    ret.append("world")
    return ret

@external
def foo6() -> DynArray[DynArray[String[32], 2], 2]:
    ret: DynArray[DynArray[String[32], 2], 2] = []
    ret.append(["hello", "world"])
    return ret
    """

    c = get_contract(code)
    assert c.foo1(["hello", "world"]) == ["hello", "world"]
    assert c.foo2([["hello", "world"]]) == [["hello", "world"]]
    assert c.foo3([["hello", "world"]]) == ["hello", "world"]
    assert c.foo4([["hello", "world"]]) == "hello"
    assert c.foo5() == ["hello", "world"]
    assert c.foo6() == [["hello", "world"]]


def test_list_output_tester_code(get_contract):
    list_output_tester_code = """
flag Foobar:
    FOO
    BAR

y: DynArray[Foobar, 2]
z: DynArray[int128, 2]

@external
def foo() -> DynArray[int128, 2]:
    return [3, 5]

@external
def goo() -> DynArray[int128, 2]:
    x: DynArray[int128, 2] = [3, 5]
    return x

@external
def hoo() -> DynArray[int128, 2]:
    self.z = [3, 5]
    return self.z

@external
def hoo1() -> DynArray[int128, 2]:
    self.z = empty(DynArray[int128, 2])
    return self.z

@external
def hoo2() -> DynArray[int128, 2]:
    return empty(DynArray[int128, 2])

@external
def hoo3() -> DynArray[int128, 2]:
    return []

@external
def hoo4() -> DynArray[int128, 2]:
    self.z = []
    return self.z

@external
def hoo5() -> DynArray[DynArray[int128, 2], 2]:
    return []

@external
def joo() -> DynArray[int128, 2]:
    self.z = [3, 5]
    x: DynArray[int128, 2] = self.z
    return x

@external
def koo() -> DynArray[DynArray[int128, 2], 2]:
    return [[1, 2], [3, 4]]

@external
def loo() -> DynArray[DynArray[int128, 2], 2]:
    x: DynArray[DynArray[int128, 2], 2] = [[1, 2], [3, 4]]
    return x

@external
def moo() -> DynArray[DynArray[int128, 2], 2]:
    x: DynArray[int128, 2] = [1,2]
    return [x, [3,4]]

@external
def noo(inp: DynArray[int128, 2]) -> DynArray[int128, 2]:
    return inp

@external
def ooo(inp: DynArray[int128, 2]) -> DynArray[int128, 2]:
    self.z = inp
    return self.z

@external
def poo(inp: DynArray[DynArray[int128, 2], 2]) -> DynArray[DynArray[int128, 2], 2]:
    return inp

@external
def qoo(inp: DynArray[int128, 2]) -> DynArray[DynArray[int128, 2], 2]:
    return [inp, [3,4]]

@external
def roo(inp: DynArray[decimal, 2]) -> DynArray[DynArray[decimal, 2], 2]:
    return [inp, [3.0, 4.0]]

@external
def soo() -> DynArray[Foobar, 2]:
    x: DynArray[Foobar, 2] = [Foobar.FOO, Foobar.BAR]
    return x

@external
def too() -> DynArray[Foobar, 2]:
    self.y = [Foobar.BAR, Foobar.FOO]
    return self.y

@external
def uoo(inp: DynArray[Foobar, 2]) -> DynArray[DynArray[Foobar, 2], 2]:
    return [inp, [Foobar.BAR, Foobar.FOO]]
    """

    c = get_contract(list_output_tester_code)
    assert c.foo() == [3, 5]
    assert c.goo() == [3, 5]
    assert c.hoo() == [3, 5]
    assert c.hoo1() == c.hoo2() == c.hoo3() == c.hoo4() == []
    assert c.hoo5() == []
    assert c.joo() == [3, 5]
    assert c.koo() == [[1, 2], [3, 4]]
    assert c.loo() == [[1, 2], [3, 4]]
    assert c.moo() == [[1, 2], [3, 4]]
    assert c.noo([]) == []
    assert c.noo([3, 5]) == [3, 5]
    assert c.ooo([]) == []
    assert c.ooo([3, 5]) == [3, 5]
    assert c.poo([]) == []
    assert c.poo([[1, 2], [3, 4]]) == [[1, 2], [3, 4]]
    assert c.qoo([1, 2]) == [[1, 2], [3, 4]]
    assert c.roo([decimal_to_int(1), decimal_to_int(2)]) == [
        [decimal_to_int(1), decimal_to_int(2)],
        [decimal_to_int(3), decimal_to_int(4)],
    ]
    assert c.soo() == [1, 2]
    assert c.too() == [2, 1]
    assert c.uoo([1, 2]) == [[1, 2], [2, 1]]

    print("Passed list output tests")


def test_array_accessor(get_contract):
    array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[int128, 4] = [0, 0, 0, 0]
    a[0] = x
    a[1] = y
    a[2] = z
    a[3] = w
    return a[0] * 1000 + a[1] * 100 + a[2] * 10 + a[3]
    """

    c = get_contract(array_accessor)
    assert c.test_array(2, 7, 1, 8) == 2718
    print("Passed basic array accessor test")


def test_two_d_array_accessor(get_contract):
    two_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[int128, 2], 2] = [[0, 0], [0, 0]]
    a[0][0] = x
    a[0][1] = y
    a[1][0] = z
    a[1][1] = w
    return a[0][0] * 1000 + a[0][1] * 100 + a[1][0] * 10 + a[1][1]
    """

    c = get_contract(two_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == 2718
    print("Passed complex array accessor test")


def test_three_d_array_accessor(get_contract):
    three_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[DynArray[int128, 2], 2], 2] = [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]
    a[0][0][0] = x
    a[0][0][1] = y
    a[0][1][0] = z
    a[0][1][1] = w
    a[1][0][0] = -x
    a[1][0][1] = -y
    a[1][1][0] = -z
    a[1][1][1] = -w
    return a[0][0][0] * 1000 + a[0][0][1] * 100 + a[0][1][0] * 10 + a[0][1][1] + \\
        a[1][1][1] * 1000 + a[1][1][0] * 100 + a[1][0][1] * 10 + a[1][0][0]
    """

    c = get_contract(three_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == -5454


def test_four_d_array_accessor(get_contract):
    four_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[DynArray[DynArray[int128, 2], 2], 2], 2] = \\
        [[[[0, 0], [0, 0]], [[0, 0], [0, 0]]], [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]]
    a[0][0][0][0] = x
    a[0][0][0][1] = y
    a[0][0][1][0] = z
    a[0][0][1][1] = w
    a[0][1][0][0] = -x
    a[0][1][0][1] = -y
    a[0][1][1][0] = -z
    a[0][1][1][1] = -w

    a[1][0][0][0] = x + 1
    a[1][0][0][1] = y + 1
    a[1][0][1][0] = z + 1
    a[1][0][1][1] = w + 1
    a[1][1][0][0] = - (x + 1)
    a[1][1][0][1] = - (y + 1)
    a[1][1][1][0] = - (z + 1)
    a[1][1][1][1] = - (w + 1)
    return a[0][0][0][0] * 1000 + a[0][0][0][1] * 100 + a[0][0][1][0] * 10 + a[0][0][1][1] + \\
        a[0][1][1][1] * 1000 + a[0][1][1][0] * 100 + a[0][1][0][1] * 10 + a[0][1][0][0] + \\
        a[1][0][0][0] * 1000 + a[1][0][0][1] * 100 + a[1][0][1][0] * 10 + a[1][0][1][1] + \\
        a[1][1][1][1] * 1000 + a[1][1][1][0] * 100 + a[1][1][0][1] * 10 + a[1][1][0][0]
    """

    c = get_contract(four_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == -10908


def test_array_negative_accessor(get_contract, assert_compile_failed):
    array_constant_negative_accessor = """
FOO: constant(int128) = -1
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[4] = [0, 0, 0, 0]
    a[0] = x
    a[1] = y
    a[2] = z
    a[3] = w
    return a[-4] * 1000 + a[-3] * 100 + a[-2] * 10 + a[FOO]
    """

    with pytest.raises(ArrayIndexException):
        compile_code(array_constant_negative_accessor)

    array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[int128, 4] = [0, 0, 0, 0]
    a[0] = x
    a[1] = y
    a[2] = z
    a[3] = w
    return a[-4] * 1000 + a[-3] * 100 + a[-2] * 10 + a[-1]
    """

    assert_compile_failed(lambda: get_contract(array_negative_accessor), ArrayIndexException)

    two_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[int128, 2], 2] = [[0, 0], [0, 0]]
    a[0][0] = x
    a[0][1] = y
    a[1][0] = z
    a[1][1] = w
    return a[-2][-2] * 1000 + a[-2][-1] * 100 + a[-1][-2] * 10 + a[-1][-1]
    """

    assert_compile_failed(lambda: get_contract(two_d_array_negative_accessor), ArrayIndexException)

    three_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[DynArray[int128, 2], 2], 2] = [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]
    a[0][0][0] = x
    a[0][0][1] = y
    a[0][1][0] = z
    a[0][1][1] = w
    a[1][0][0] = -x
    a[1][0][1] = -y
    a[1][1][0] = -z
    a[1][1][1] = -w
    return a[-2][-2][-2] * 1000 + a[-2][-2][-1] * 100 + a[-2][-1][-2] * 10 + a[-2][-1][-1] + \\
        a[-1][-1][-1] * 1000 + a[-1][-1][-2] * 100 + a[-1][-2][-1] * 10 + a[-1][-2][-2]
    """

    assert_compile_failed(
        lambda: get_contract(three_d_array_negative_accessor), ArrayIndexException
    )

    four_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: DynArray[DynArray[DynArray[DynArray[int128, 2], 2], 2], 2] = \\
        [[[[0, 0], [0, 0]], [[0, 0], [0, 0]]], [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]]
    a[0][0][0][0] = x
    a[0][0][0][1] = y
    a[0][0][1][0] = z
    a[0][0][1][1] = w
    a[0][1][0][0] = -x
    a[0][1][0][1] = -y
    a[0][1][1][0] = -z
    a[0][1][1][1] = -w

    a[1][0][0][0] = x + 1
    a[1][0][0][1] = y + 1
    a[1][0][1][0] = z + 1
    a[1][0][1][1] = w + 1
    a[1][1][0][0] = - (x + 1)
    a[1][1][0][1] = - (y + 1)
    a[1][1][1][0] = - (z + 1)
    a[1][1][1][1] = - (w + 1)
    return a[-2][-2][-2][-2] * 1000 + a[-2][-2][-2][-1] * 100 + \\
        a[-2][-2][-1][-2] * 10 + a[-2][-2][-1][-1] + \\
        a[-2][-1][-1][-1] * 1000 + a[-2][-1][-1][-2] * \\
        100 + a[-2][-1][-2][-1] * 10 + a[-2][-1][-2][-2] + \\
        a[-1][-2][-2][-2] * 1000 + a[-1][-2][-2][-1] * \\
        100 + a[-1][-2][-1][-2] * 10 + a[-1][-2][-1][-1] + \\
        a[-1][-1][-1][-1] * 1000 + a[-1][-1][-1][-2] * \\
        100 + a[-1][-1][-2][-1] * 10 + a[-1][-1][-2][-2]
    """

    assert_compile_failed(lambda: get_contract(four_d_array_negative_accessor), ArrayIndexException)


@pytest.mark.parametrize(
    "type,values,false_value",
    [
        ("uint256", [3, 7], 4),
        (
            "address",
            [
                "0x0000000000000000000000000000000000000012",
                "0x0000000000000000000000000000000000000024",
            ],
            "0x0000000000000000000000000000000000000013",
        ),
        ("bool", [True, True], False),
        (
            "bytes32",
            [
                "0x0000000000000000000000000000000000000000000000000000000080ac58cd",
                "0x0000000000000000000000000000000000000000000000000000000080ac58ce",
            ],
            "0x0000000000000000000000000000000000000000000000000000000080ac58cf",
        ),
    ],
)
def test_member_in_list(get_contract, type, values, false_value):
    code = f"""
@external
def check(a: {type}) -> bool:
    x: DynArray[{type}, 2] = [{values[0]}, {values[1]}]
    return a in x
    """
    c = get_contract(code)
    assert c.check(values[0]) is True
    assert c.check(values[1]) is True
    assert c.check(false_value) is False


@pytest.mark.parametrize("type_", ("uint256", "bytes32", "address"))
def test_member_in_empty_list(get_contract, type_):
    code = f"""
@external
def check_in(s: uint128) -> bool:
    a: {type_} = convert(s, {type_})
    x: DynArray[{type_}, 2] = []
    return a in x

@external
def check_not_in(s: uint128) -> bool:
    a: {type_} = convert(s, {type_})
    x: DynArray[{type_}, 2] = []
    return a not in x
    """
    c = get_contract(code)
    for s in (0, 1, 2, 3):
        assert c.check_in(s) is False
        assert c.check_not_in(s) is True


@pytest.mark.parametrize(
    "type,values,false_values",
    [
        ("uint256", [[3, 7], [9, 11]], [4, 10]),
        ("bool", [[True, True], [False, False]], [False, True]),
    ],
)
def test_member_in_nested_list(get_contract, type, values, false_values):
    code = f"""
@external
def check1(a: {type}) -> bool:
    x: DynArray[DynArray[{type}, 2], 2] = {values}
    return a in x[0]

@external
def check2(a: {type}) -> bool:
    x: DynArray[DynArray[{type}, 2], 2] = {values}
    return a in x[1]
    """
    c = get_contract(code)
    assert c.check1(values[0][0]) is True
    assert c.check1(values[0][1]) is True
    assert c.check1(false_values[0]) is False

    assert c.check2(values[1][0]) is True
    assert c.check2(values[1][1]) is True
    assert c.check2(false_values[1]) is False


def test_member_in_nested_address_list(get_contract):
    code = """
@external
def check1(a: address) -> bool:
    x: DynArray[DynArray[address, 2], 2] = [
        [
            0x0000000000000000000000000000000000000012,
            0x0000000000000000000000000000000000000024,
        ],
        [
            0x0000000000000000000000000000000000000036,
            0x0000000000000000000000000000000000000048,
        ],
    ]
    return a in x[0]

@external
def check2(a: address) -> bool:
    x: DynArray[DynArray[address, 2], 2] = [
        [
            0x0000000000000000000000000000000000000012,
            0x0000000000000000000000000000000000000024,
        ],
        [
            0x0000000000000000000000000000000000000036,
            0x0000000000000000000000000000000000000048,
        ],
    ]
    return a in x[1]
    """
    c = get_contract(code)
    assert c.check1("0x0000000000000000000000000000000000000012") is True
    assert c.check1("0x0000000000000000000000000000000000000024") is True
    assert c.check1("0x0000000000000000000000000000000000000036") is False

    assert c.check2("0x0000000000000000000000000000000000000036") is True
    assert c.check2("0x0000000000000000000000000000000000000048") is True
    assert c.check2("0x0000000000000000000000000000000000000024") is False


def test_member_in_nested_bytes32_list(get_contract):
    code = """
@external
def check1(a: bytes32) -> bool:
    x: DynArray[DynArray[bytes32, 2], 2] = [
        [
            0x0000000000000000000000000000000000000000000000000000000080ac58ca,
            0x0000000000000000000000000000000000000000000000000000000080ac58cb,
        ],
        [
            0x0000000000000000000000000000000000000000000000000000000080ac58cc,
            0x0000000000000000000000000000000000000000000000000000000080ac58cd,
        ],
    ]
    return a in x[0]

@external
def check2(a: bytes32) -> bool:
    x: DynArray[DynArray[bytes32, 2], 2] = [
        [
            0x0000000000000000000000000000000000000000000000000000000080ac58ca,
            0x0000000000000000000000000000000000000000000000000000000080ac58cb,
        ],
        [
            0x0000000000000000000000000000000000000000000000000000000080ac58cc,
            0x0000000000000000000000000000000000000000000000000000000080ac58cd,
        ],
    ]
    return a in x[1]
    """
    c = get_contract(code)
    assert c.check1("0x0000000000000000000000000000000000000000000000000000000080ac58ca") is True
    assert c.check1("0x0000000000000000000000000000000000000000000000000000000080ac58cb") is True
    assert c.check1("0x0000000000000000000000000000000000000000000000000000000080ac58cc") is False

    assert c.check2("0x0000000000000000000000000000000000000000000000000000000080ac58cc") is True
    assert c.check2("0x0000000000000000000000000000000000000000000000000000000080ac58cd") is True
    assert c.check2("0x0000000000000000000000000000000000000000000000000000000080ac58ca") is False


def test_member_in_updated_list(get_contract):
    code = """
@external
def foo() -> bool:
    xs: DynArray[uint256, 3] = [2, 2, 2]
    xs = [1, 1]
    y: uint256 = 2
    return y in xs
    """
    c = get_contract(code)
    assert c.foo() is False


def test_member_in_updated_nested_list(get_contract):
    code = """
@external
def foo() -> bool:
    xs: DynArray[DynArray[DynArray[uint256, 3], 3], 3] = [
        [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
        [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
        [[2, 2, 2], [2, 2, 2], [2, 2, 2]],
    ]
    xs = [
        [[1, 1], [1, 1], [1, 1]],
        [[1, 1], [1, 1], [1, 1]],
        [[1, 1], [1, 1], [1, 1]],
    ]
    y: uint256 = 2
    return y in xs[0][0] or y in xs[0][1] or y in xs[0][2] or \\
        y in xs[1][0] or y in xs[1][1] or y in xs[1][2] or \\
        y in xs[2][0] or y in xs[2][1] or y in xs[2][2]
    """
    c = get_contract(code)
    assert c.foo() is False


def test_member_in_list_lhs_side_effects(get_contract):
    code = """
_counter: uint256

@internal
def counter() -> uint256:
    self._counter = 1
    return self._counter

@external
def bar() -> bool:
    x: DynArray[uint256, 4] = [2, 2, 2, 2]
    return self.counter() in x
    """
    c = get_contract(code)
    assert c.bar() is False


def test_member_in_nested_list_lhs_side_effects(get_contract):
    code = """
_counter: uint256

@internal
def counter() -> uint256:
    self._counter = 1
    return self._counter

@external
def bar() -> bool:
    x: DynArray[DynArray[DynArray[uint256, 4], 4], 4] = [
        [[2, 2, 2, 2], [2, 2, 2, 2], [2, 2, 2, 2]],
        [[2, 2, 2, 2], [2, 2, 2, 2], [2, 2, 2, 2]],
        [[2, 2, 2, 2], [2, 2, 2, 2], [2, 2, 2, 2]],
    ]
    return self.counter() in x[0][0]
    """
    c = get_contract(code)
    assert c.bar() is False


def test_member_in_list_rhs_side_effects(get_contract):
    code = """
counter: uint256

@internal
def foo() -> DynArray[uint256, 3]:
    self.counter += 1
    return [0,0,0]

@external
def bar() -> uint256:
    self.counter = 0
    t: bool = self.counter in self.foo()
    return self.counter
    """
    c = get_contract(code)
    assert c.bar() == 1


def test_member_in_nested_list_rhs_side_effects(get_contract):
    code = """
counter: uint256

@internal
def foo() -> DynArray[DynArray[DynArray[uint256, 3], 3], 3]:
    self.counter += 1
    return [
        [[0,0,0], [0,0,0], [0,0,0]],
        [[0,0,0], [0,0,0], [0,0,0]],
        [[0,0,0], [0,0,0], [0,0,0]]
    ]

@external
def bar() -> uint256:
    self.counter = 0
    t: bool = self.counter in self.foo()[0][0]
    return self.counter
    """
    c = get_contract(code)
    assert c.bar() == 1


def test_returns_lists(get_contract):
    code = """
@external
def test_array_num_return() -> DynArray[DynArray[int128, 2], 2]:
    a: DynArray[DynArray[int128, 2], 2] = [empty(DynArray[int128, 2]), [3, 4]]
    return a

@external
def test_array_decimal_return1() -> DynArray[DynArray[decimal, 2], 2]:
    a: DynArray[DynArray[decimal, 2], 2] = [[1.0], [3.0, 4.0]]
    return a

@external
def test_array_decimal_return2() -> DynArray[DynArray[decimal, 2], 2]:
    return [[1.0, 2.0]]

@external
def test_array_decimal_return3() -> DynArray[DynArray[decimal, 2], 2]:
    a: DynArray[DynArray[decimal, 2], 2] = [[1.0, 2.0], [3.0]]
    return a
"""

    c = get_contract(code)
    assert c.test_array_num_return() == [[], [3, 4]]
    assert c.test_array_decimal_return1() == [
        [decimal_to_int(1)],
        [decimal_to_int(3), decimal_to_int(4)],
    ]
    assert c.test_array_decimal_return2() == [[decimal_to_int(1), decimal_to_int(2)]]
    assert c.test_array_decimal_return3() == [
        [decimal_to_int(1), decimal_to_int(2)],
        [decimal_to_int(3)],
    ]


@pytest.mark.venom_xfail(raises=StackTooDeep, reason="stack scheduler regression")
def test_mult_list(get_contract):
    code = """
nest3: DynArray[DynArray[DynArray[uint256, 2], 2], 2]
nest4: DynArray[DynArray[DynArray[DynArray[uint256, 2], 2], 2], 2]

@external
def test_multi3_1() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    l: DynArray[DynArray[DynArray[uint256, 2], 2], 2] = [[[0, 0], [0, 4]], [[0, 7], [0, 123]]]
    self.nest3 = l
    return self.nest3

@external
def test_multi3_2() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    l: DynArray[DynArray[DynArray[uint256, 2], 2], 2] = [[[0, 0], [0, 4]], [[0, 7], [0, 123]]]
    self.nest3 = l
    l = self.nest3
    return l

@external
def test_multi4_1() -> DynArray[DynArray[DynArray[DynArray[uint256, 2], 2], 2], 2]:
    l: DynArray[DynArray[DynArray[DynArray[uint256, 2], 2], 2], 2] = [[[[1, 0], [0, 4]], [[0, 0], [0, 0]]], [[[444, 0], [0, 0]],[[1, 0], [0, 222]]]]  # noqa: E501
    self.nest4 = l
    l = self.nest4
    return l

@external
def test_multi4_2() -> DynArray[DynArray[DynArray[DynArray[uint256, 2], 2], 2], 2]:
    l: DynArray[DynArray[DynArray[DynArray[uint256, 2], 2], 2], 2] = [[[[1, 0], [0, 4]], [[0, 0], [0, 0]]], [[[444, 0], [0, 0]],[[1, 0], [0, 222]]]]  # noqa: E501
    self.nest4 = l
    return self.nest4
    """

    c = get_contract(code)

    nest3 = [[[0, 0], [0, 4]], [[0, 7], [0, 123]]]
    assert c.test_multi3_1() == nest3
    assert c.test_multi3_2() == nest3
    nest4 = [[[[1, 0], [0, 4]], [[0, 0], [0, 0]]], [[[444, 0], [0, 0]], [[1, 0], [0, 222]]]]
    assert c.test_multi4_1() == nest4
    assert c.test_multi4_2() == nest4


def test_uint256_accessor(get_contract, tx_failed):
    code = """
@external
def bounds_check_uint256(xs: DynArray[uint256, 3], ix: uint256) -> uint256:
    return xs[ix]
    """
    c = get_contract(code)
    with tx_failed():
        c.bounds_check_uint256([], 0)

    assert c.bounds_check_uint256([1], 0) == 1
    with tx_failed():
        c.bounds_check_uint256([1], 1)

    assert c.bounds_check_uint256([1, 2, 3], 0) == 1
    assert c.bounds_check_uint256([1, 2, 3], 2) == 3
    with tx_failed():
        c.bounds_check_uint256([1, 2, 3], 3)

    # TODO do bounds checks for nested darrays


@pytest.mark.parametrize("list_", ([], [11], [11, 12], [11, 12, 13]))
def test_dynarray_len(get_contract, tx_failed, list_):
    code = """
@external
def darray_len(xs: DynArray[uint256, 3]) -> uint256:
    return len(xs)
    """

    c = get_contract(code)
    assert c.darray_len(list_) == len(list_)


def test_dynarray_too_large(get_contract, tx_failed):
    code = """
@external
def darray_len(xs: DynArray[uint256, 3]) -> uint256:
    return len(xs)
    """

    c = get_contract(code)
    with tx_failed():
        c.darray_len([1, 2, 3, 4])


def test_int128_accessor(get_contract, tx_failed):
    code = """
@external
def bounds_check_int128(ix: int128) -> uint256:
    xs: DynArray[uint256, 3] = [1,2,3]
    return xs[ix]
    """
    c = get_contract(code)
    assert c.bounds_check_int128(0) == 1
    assert c.bounds_check_int128(2) == 3
    with tx_failed():
        c.bounds_check_int128(3)
    with tx_failed():
        c.bounds_check_int128(-1)


def test_index_exception(get_contract, assert_compile_failed):
    code = """
@external
def fail() -> uint256:
    xs: DynArray[uint256, 3] = [1,2,3]
    return xs[3]
    """
    assert_compile_failed(lambda: get_contract(code), ArrayIndexException)

    code = """
@external
def fail() -> uint256:
    xs: DynArray[uint256, 3] = [1,2,3]
    return xs[-1]
    """
    assert_compile_failed(lambda: get_contract(code), ArrayIndexException)


def test_compile_time_bounds_check(get_contract, assert_compile_failed):
    code = """
@external
def parse_list_fail():
    xs: DynArray[uint256, 3] = [2**256, 1, 3]
    pass
    """
    assert_compile_failed(lambda: get_contract(code), OverflowException)


def test_2d_array_input_1(get_contract):
    code = """
@internal
def test_input(
    arr: DynArray[DynArray[int128, 2], 1], i: int128
) -> (DynArray[DynArray[int128, 2], 1], int128):
    return arr, i

@external
def test_values(
    arr: DynArray[DynArray[int128, 2], 1], i: int128
) -> (DynArray[DynArray[int128, 2], 1], int128):
    return self.test_input(arr, i)
    """

    c = get_contract(code)
    assert c.test_values([[1, 2]], 3) == ([[1, 2]], 3)


def test_2d_array_input_2(get_contract):
    code = """
@internal
def test_input(
    arr: DynArray[DynArray[int128, 2], 3],
    s: String[10]
) -> (DynArray[DynArray[int128, 2], 3], String[10]):
    return arr, s

@external
def test_values(
    arr: DynArray[DynArray[int128, 2], 3],
    s: String[10]
) -> (DynArray[DynArray[int128, 2], 3], String[10]):
    return self.test_input(arr, s)
    """

    c = get_contract(code)
    assert c.test_values([[1, 2], [3, 4], [5, 6]], "abcdef") == ([[1, 2], [3, 4], [5, 6]], "abcdef")


def test_nested_index_of_returned_array(get_contract):
    code = """
@internal
def inner() -> (int128, int128):
    return 1,2

@external
def outer() -> DynArray[int128, 2]:
    return [333, self.inner()[0]]
    """

    c = get_contract(code)
    assert c.outer() == [333, 1]


def test_nested_calls_inside_arrays(get_contract):
    code = """
@internal
def _foo(a: uint256, b: DynArray[uint256, 2]) -> (uint256, uint256, uint256, uint256, uint256):
    return 1, a, b[0], b[1], 5

@internal
def _foo2() -> uint256:
    a: DynArray[uint256, 10] = [6,7,8,9,10,11,12,13,15,16]
    return 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return self._foo(2, [3, self._foo2()])
    """

    c = get_contract(code)
    assert c.foo() == (1, 2, 3, 4, 5)


def test_nested_calls_inside_arrays_with_index_access(get_contract):
    code = """
@internal
def _foo(
    a: DynArray[uint256, 2],
    b: DynArray[uint256, 2]
) -> (uint256, uint256, uint256, uint256, uint256):
    return a[1]-b[0], 2, a[0]-b[1], 8-b[1], 5

@internal
def _foo2() -> (uint256, uint256):
    a: DynArray[uint256, 10] = [6,7,8,9,10,11,12,13,15,16]
    return a[6], 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return self._foo([7, self._foo2()[0]], [11, self._foo2()[1]])
    """

    c = get_contract(code)
    assert c.foo() == (1, 2, 3, 4, 5)


append_pop_tests = [
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    for x: uint256 in xs:
        self.my_array.append(x)
    return self.my_array
    """,
        lambda xs: xs,
    ),
    (
        """
my_array: DynArray[uint256, 5]
some_var: uint256
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    for x: uint256 in xs:
        self.some_var = x
        # test that typechecker for append args works
        self.my_array.append(self.some_var)
    return self.my_array
    """,
        lambda xs: xs,
    ),
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    for x: uint256 in xs:
        self.my_array.append(x)
    for x: uint256 in xs:
        self.my_array.pop()
    return self.my_array
    """,
        lambda xs: [],
    ),
    # check order of evaluation.
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 5]) -> (DynArray[uint256, 5], uint256):
    for x: uint256 in xs:
        self.my_array.append(x)
    return self.my_array, self.my_array.pop()
    """,
        lambda xs: None if len(xs) == 0 else (xs, xs[-1]),
    ),
    # check order of evaluation.
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 5]) -> (uint256, DynArray[uint256, 5]):
    for x: uint256 in xs:
        self.my_array.append(x)
    return self.my_array.pop(), self.my_array
    """,
        lambda xs: None if len(xs) == 0 else (xs[-1], xs[:-1]),
    ),
    # test memory arrays
    (
        """
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    ys: DynArray[uint256, 5] = []
    i: uint256 = 0
    for x: uint256 in xs:
        if i >= len(xs) - 1:
            break
        ys.append(x)
        i += 1

    return ys
    """,
        lambda xs: xs[:-1],
    ),
    # check overflow
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 6]) -> DynArray[uint256, 5]:
    for x: uint256 in xs:
        self.my_array.append(x)
    return self.my_array
    """,
        lambda xs: None if len(xs) > 5 else xs,
    ),
    # pop to 0 elems
    (
        """
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    ys: DynArray[uint256, 5] = []
    for x: uint256 in xs:
        ys.append(x)
    for x: uint256 in xs:
        ys.pop()
    return ys
    """,
        lambda xs: [],
    ),
    # check underflow
    (
        """
@external
def foo(xs: DynArray[uint256, 5]) -> DynArray[uint256, 5]:
    ys: DynArray[uint256, 5] = []
    for x: uint256 in xs:
        ys.append(x)
    for x: uint256 in xs:
        ys.pop()
    ys.pop()  # fail
    return ys
    """,
        lambda xs: None,
    ),
    # check underflow
    (
        """
my_array: DynArray[uint256, 5]
@external
def foo(xs: DynArray[uint256, 5]) -> uint256:
    return self.my_array.pop()
    """,
        lambda xs: None,
    ),
]


@pytest.mark.parametrize("subtyp", ["uint8", "int128", "uint256"])
def test_append_literal(get_contract, subtyp):
    data = [1, 2, 3]
    if subtyp == "int128":
        data = [-1, 2, 3]
    code = f"""
@external
def foo() -> DynArray[{subtyp}, 3]:
    x: DynArray[{subtyp}, 3] = []
    x.append({data[0]})
    x.append({data[1]})
    x.append({data[2]})
    return x
    """
    c = get_contract(code)
    assert c.foo() == data


@pytest.mark.parametrize("subtyp,lit", [("uint8", 256), ("uint256", -1), ("int128", 2**127)])
def test_append_invalid_literal(get_contract, assert_compile_failed, subtyp, lit):
    code = f"""
@external
def foo() -> DynArray[{subtyp}, 3]:
    x: DynArray[{subtyp}, 3] = []
    x.append({lit})
    return x
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


invalid_appends_pops = [
    (
        """
@external
def foo() -> DynArray[uint256, 3]:
    x: DynArray[uint256, 3] = []
    x.append()
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo() -> DynArray[uint256, 3]:
    x: DynArray[uint256, 3] = []
    x.append(1,2)
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo() -> DynArray[uint256, 3]:
    x: DynArray[uint256, 3] = []
    x.pop(1)
    """,
        ArgumentException,
    ),
    (
        """
@external
def foo(x: DynArray[uint256, 3]) -> DynArray[uint256, 3]:
    x.append(1)
    return x
    """,
        ImmutableViolation,
    ),
    (
        """
foo: DynArray[uint256, 3]
@external
@view
def bar() -> DynArray[uint256, 3]:
    self.foo.append(1)
    return self.foo
    """,
        StateAccessViolation,
    ),
]


@pytest.mark.parametrize("code,exception_type", invalid_appends_pops)
def test_invalid_append_pop(get_contract, assert_compile_failed, code, exception_type):
    assert_compile_failed(lambda: get_contract(code), exception_type)


@pytest.mark.parametrize("code,check_result", append_pop_tests)
# TODO change this to fuzz random data
@pytest.mark.parametrize("test_data", [[1, 2, 3, 4, 5][:i] for i in range(6)])
def test_append_pop(get_contract, tx_failed, code, check_result, test_data):
    c = get_contract(code)
    expected_result = check_result(test_data)
    if expected_result is None:
        # None is sentinel to indicate txn should revert
        with tx_failed():
            c.foo(test_data)
    else:
        assert c.foo(test_data) == expected_result


append_pop_complex_tests: list[tuple[str, Callable[[Any], Any]]] = [
    (
        """
@external
def foo(x: {typ}) -> DynArray[{typ}, 2]:
    ys: DynArray[{typ}, 1] = []
    ys.append(x)
    return ys
    """,
        lambda x: [x],
    ),
    (
        """
my_array: DynArray[{typ}, 1]
@external
def foo(x: {typ}) -> DynArray[{typ}, 2]:
    self.my_array.append(x)
    self.my_array.append(x)  # fail
    return self.my_array
    """,
        lambda x: None,
    ),
    (
        """
my_array: DynArray[{typ}, 5]
@external
def foo(x: {typ}) -> (DynArray[{typ}, 5], {typ}):
    self.my_array.append(x)
    return self.my_array, self.my_array.pop()
    """,
        lambda x: ([x], x),
    ),
    (
        """
my_array: DynArray[{typ}, 5]
@external
def foo(x: {typ}) -> ({typ}, DynArray[{typ}, 5]):
    self.my_array.append(x)
    return self.my_array.pop(), self.my_array
    """,
        lambda x: (x, []),
    ),
    (
        """
my_array: DynArray[{typ}, 5]
@external
def foo(x: {typ}) -> {typ}:
    return self.my_array.pop()
    """,
        lambda x: None,
    ),
]


@pytest.mark.parametrize("code_template,check_result", append_pop_complex_tests)
@pytest.mark.parametrize(
    "subtype",
    ["uint256[3]", "DynArray[uint256,3]", "DynArray[uint8, 4]", "Foo", "DynArray[Foobar, 3]"],
)
# TODO change this to fuzz random data
def test_append_pop_complex(get_contract, tx_failed, code_template, check_result, subtype):
    code = code_template.format(typ=subtype)
    test_data = [1, 2, 3]
    if subtype == "Foo":
        test_data = tuple(test_data)
        struct_def = """
struct Foo:
    x: uint256
    y: uint256
    z: uint256
        """
        code = struct_def + "\n" + code
    elif subtype == "DynArray[Foobar, 3]":
        flag_def = """
flag Foobar:
    FOO
    BAR
    BAZ
        """
        code = flag_def + "\n" + code
        test_data = [2 ** (i - 1) for i in test_data]

    c = get_contract(code)
    expected_result = check_result(test_data)
    if expected_result is None:
        # None is sentinel to indicate txn should revert
        with tx_failed():
            c.foo(test_data)
    else:
        assert c.foo(test_data) == expected_result


def test_so_many_things_you_should_never_do(get_contract):
    code = """
@internal
def _foo(a: DynArray[uint256, 2], b: DynArray[uint256, 2]) -> DynArray[uint256, 5]:
    return [a[1]-b[0], 2, a[0]-b[1], 8-b[1], 5]

@internal
def _foo2() -> (uint256, uint256):
    b: DynArray[uint256, 2] = [5, 8]
    a: DynArray[uint256, 10] = [6,7,8,9,10,11,12,13,self._foo([44,b[0]],b)[4],16]
    return a[6], 4

@external
def foo() -> (uint256, DynArray[uint256, 3], DynArray[uint256, 2]):
    x: DynArray[uint256, 3] = [
        1,
        14-self._foo2()[0],
        self._foo([7,self._foo2()[0]], [11,self._foo2()[1]])[2]
    ]
    return 666, x, [88, self._foo2()[0]]
    """
    c = get_contract(code)
    assert c.foo() == (666, [1, 2, 3], [88, 12])


def test_list_of_structs_arg(get_contract):
    code = """
flag Foobar:
    FOO
    BAR

struct Foo:
    x: uint256
    y: uint256
    z: Foobar

@external
def bar(_baz: DynArray[Foo, 3]) -> uint256:
    sum: uint256 = 0
    for i: uint256 in range(3):
        e: Foobar = _baz[i].z
        f: uint256 = convert(e, uint256)
        sum += _baz[i].x * _baz[i].y + f
    return sum
    """
    c = get_contract(code)
    c_input = [[x, y, 1] for x, y in zip(range(3), range(3))]
    assert c.bar(c_input) == 8  # (0 * 0 + 1) + (1 * 1 + 1) + (2 * 2 + 1)


def test_list_of_structs_arg_with_dynamic_type(get_contract):
    code = """
struct Foo:
    x: uint256
    _msg: String[32]

@external
def bar(_baz: DynArray[Foo, 3]) -> String[96]:
    return concat(_baz[0]._msg, _baz[1]._msg, _baz[2]._msg)
    """
    c = get_contract(code)
    c_input = [[i, msg] for i, msg in enumerate(("Hello ", "world", "!!!!"))]
    assert c.bar(c_input) == "Hello world!!!!"


def test_list_of_structs_lists_with_nested_lists(get_contract, tx_failed):
    code = """
struct Bar:
    a: DynArray[uint8[2], 2]

@external
def foo(x: uint8) -> uint8:
    b: DynArray[Bar[2], 2] = [
        [
            Bar(a=[[x, x + 1], [x + 2, x + 3]]),
            Bar(a=[[x + 4, x +5], [x + 6, x + 7]])
        ],
        [
            Bar(a=[[x + 8, x + 9], [x + 10, x + 11]]),
            Bar(a=[[x + 12, x + 13], [x + 14, x + 15]])
        ],
    ]
    return b[0][0].a[0][0] + b[0][1].a[1][1] + b[1][0].a[0][1] + b[1][1].a[1][0]
    """
    c = get_contract(code)
    assert c.foo(17) == 98
    with tx_failed():
        c.foo(241)


def test_list_of_nested_struct_arrays(get_contract):
    code = """
struct Ded:
    a: uint256[3]
    b: bool

struct Foo:
    c: uint256
    d: uint256
    e: Ded

struct Bar:
    f: DynArray[Foo, 3]
    g: DynArray[uint256, 3]

@external
def bar(_bar: DynArray[Bar, 3]) -> uint256:
    sum: uint256 = 0
    for i: uint256 in range(3):
        sum += _bar[i].f[0].e.a[0] * _bar[i].f[1].e.a[1]
    return sum
    """
    c = get_contract(code)
    c_input = [
        ((tuple([(123, 456, ([i, i + 1, i + 2], False))] * 3)), [9, 8, 7]) for i in range(1, 4)
    ]

    assert c.bar(c_input) == 20


def test_2d_list_of_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256

@external
def foo(x: DynArray[DynArray[Bar, 2], 2]) -> uint256:
    return x[0][0].a + x[1][1].b
    """
    c = get_contract(code)
    c_input = [([i, i * 2], [i * 3, i * 4]) for i in range(1, 3)]
    assert c.foo(c_input) == 9


def test_3d_list_of_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256

@external
def foo(x: DynArray[DynArray[DynArray[Bar, 2], 2], 2]) -> uint256:
    return x[0][0][0].a + x[1][1][1].b
    """
    c = get_contract(code)
    c_input = [([([i, i * 2], [i * 3, i * 4]) for i in range(1, 3)])] * 2
    assert c.foo(c_input) == 9


def test_list_of_static_list(get_contract):
    code = """
@external
def bar(x: int128) -> DynArray[int128[2], 2]:
    a: DynArray[int128[2], 2] = [[x, x * 2], [x * 3, x * 4]]
    return a

@external
def foo(x: int128) -> int128:
    a: DynArray[int128[2], 2] = [[x, x * 2], [x * 3, x * 4]]
    return a[0][0] * a[1][1]
    """
    c = get_contract(code)
    assert c.bar(7) == [[7, 14], [21, 28]]
    assert c.foo(7) == 196


def test_list_of_static_nested_list(get_contract):
    code = """
@external
def bar(x: int128) -> DynArray[int128[2][2], 2]:
    a: DynArray[int128[2][2], 2] = [
        [[x, x * 2], [x * 3, x * 4]],
        [[x * 5, x * 6], [x * 7, x * 8]],
    ]
    return a

@external
def foo(x: int128) -> int128:
    a: DynArray[int128[2][2], 2] = [
        [[x, x * 2], [x * 3, x * 4]],
        [[x * 5, x * 6], [x * 7, x * 8]],
    ]
    return a[0][0][0] * a[1][1][1]
    """
    c = get_contract(code)
    assert c.bar(7) == [[[7, 14], [21, 28]], [[35, 42], [49, 56]]]
    assert c.foo(7) == 392


def test_struct_of_lists(get_contract):
    code = """
struct Foo:
    a1: DynArray[uint256, 2]
    a2: DynArray[DynArray[uint256, 2], 2]
    a3: DynArray[DynArray[DynArray[uint256, 2], 2], 2]

@internal
def _foo() -> DynArray[uint256, 2]:
    return [3, 7]

@internal
def _foo2() -> DynArray[DynArray[uint256, 2], 2]:
    y: DynArray[uint256, 2] = self._foo()
    z: DynArray[uint256, 2] = [y[1], y[0]]
    return [y, z]

@internal
def _foo3() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    y: DynArray[DynArray[uint256, 2], 2] = self._foo2()
    z: DynArray[DynArray[uint256, 2], 2] = [y[1], y[0]]
    return [y, z]

@external
def bar() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    foo: Foo = Foo(
        a1=self._foo(),
        a2=self._foo2(),
        a3=self._foo3(),
    )
    return foo.a3
    """
    c = get_contract(code)
    assert c.bar() == [[[3, 7], [7, 3]], [[7, 3], [3, 7]]]


def test_struct_of_lists_2(get_contract):
    code = """
struct Foo:
    b: Bytes[32]
    da: DynArray[int128, 5]
    sa: int128[5]
    some_int: int128

@internal
def _foo(x: int128) -> Foo:
    f: Foo = Foo(
        b=b"hello",
        da=[x, x * 2],
        sa=[x + 1, x + 2, x + 3, x + 4, x + 5],
        some_int=x - 1
    )
    return f

@external
def bar(x: int128) -> DynArray[int128, 5]:
    f: Foo = self._foo(x)
    return f.da
    """
    c = get_contract(code)
    assert c.bar(7) == [7, 14]


def test_struct_of_lists_3(get_contract):
    code = """
struct Foo:
    a: DynArray[int128, 3]
    b: DynArray[address, 3]
    c: DynArray[bool, 3]

@internal
def _foo(x: int128) -> Foo:
    f: Foo = Foo(
        a=[x, x * 2],
        b=[0x0000000000000000000000000000000000000012],
        c=[False, True, False]
    )
    return f

@external
def bar(x: int128) -> DynArray[int128, 3]:
    f: Foo = self._foo(x)
    return f.a
    """
    c = get_contract(code)
    assert c.bar(7) == [7, 14]


def test_nested_struct_of_lists(get_contract, assert_compile_failed, optimize):
    code = """
struct nestedFoo:
    a1: DynArray[DynArray[DynArray[uint256, 2], 2], 2]

struct Foo:
    b1: DynArray[DynArray[DynArray[nestedFoo, 2], 2], 2]

@internal
def _foo() -> nestedFoo:
    return nestedFoo(a1=[
        [[3, 7], [7, 3]],
        [[7, 3], [3, 7]],
    ])

@internal
def _foo2() -> Foo:
    _nF1: nestedFoo = self._foo()
    return Foo(b1=[[[_nF1, _nF1], [_nF1, _nF1]], [[_nF1, _nF1], [_nF1, _nF1]]])

@internal
def _foo3(f: Foo) -> Foo:
    new_f: Foo = f
    new_f.b1[0][1][0].a1[0][0] = [0, 0]
    new_f.b1[1][0][0].a1[0][1] = [0, 0]
    new_f.b1[1][1][0].a1[1][1] = [0, 0]
    return new_f

@external
def bar() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    foo: Foo = self._foo2()
    return self._foo3(foo).b1[1][1][0].a1

@external
def bar2() -> uint256:
    foo: Foo = self._foo2()
    newFoo: Foo = self._foo3(foo)
    return newFoo.b1[1][1][0].a1[1][1][0] + \\
        newFoo.b1[1][0][0].a1[0][1][1] + \\
        newFoo.b1[0][1][0].a1[0][0][0]
    """
    c = get_contract(code)
    assert c.bar() == [[[3, 7], [7, 3]], [[7, 3], [0, 0]]]
    assert c.bar2() == 0


def test_tuple_of_lists(get_contract):
    code = """
@internal
def _foo() -> DynArray[uint256, 2]:
    return [3, 7]

@internal
def _foo2() -> DynArray[DynArray[uint256, 2], 2]:
    y: DynArray[uint256, 2] = self._foo()
    z: DynArray[uint256, 2] = [y[1], y[0]]
    return [y, z]

@internal
def _foo3() -> DynArray[DynArray[DynArray[uint256, 2], 2], 2]:
    y: DynArray[DynArray[uint256, 2], 2] = self._foo2()
    z: DynArray[DynArray[uint256, 2], 2] = [y[1], y[0]]
    return [y, z]

@internal
def _foo4() -> (DynArray[DynArray[uint256, 2], 2], DynArray[DynArray[DynArray[uint256, 2], 2], 2]):
    return (self._foo2(), self._foo3())

@external
def bar() -> uint256:
    a: DynArray[DynArray[uint256, 2], 2] = [[0, 0], [0, 0]]
    b: DynArray[DynArray[DynArray[uint256, 2], 2], 2] = [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]
    a, b = self._foo4()
    return a[0][0] * b[1][0][1] + a[1][0] * b[0][1][0]
    """
    c = get_contract(code)
    assert c.bar() == 58


def test_constant_list(get_contract, tx_failed):
    some_good_primes = [5.0, 11.0, 17.0, 29.0, 37.0, 41.0]
    code = f"""
MY_LIST: constant(DynArray[decimal, 6]) = {some_good_primes}
@external
def ix(i: uint256) -> decimal:
    return MY_LIST[i]
    """
    c = get_contract(code)
    for i, p in enumerate(some_good_primes):
        assert c.ix(i) == decimal_to_int(p)
    # assert oob
    with tx_failed():
        c.ix(len(some_good_primes) + 1)


def test_public_dynarray(get_contract):
    code = """
my_list: public(DynArray[uint256, 5])
@deploy
def __init__():
    self.my_list = [1,2,3]
    """
    c = get_contract(code)

    for i, t in enumerate([1, 2, 3]):
        assert c.my_list(i) == t


def test_nested_public_dynarray(get_contract):
    code = """
my_list: public(DynArray[DynArray[uint256, 5], 5])
@deploy
def __init__():
    self.my_list = [[1,2,3]]
    """
    c = get_contract(code)

    for i, l in enumerate([[1, 2, 3]]):
        for j, t in enumerate(l):
            assert c.my_list(i, j) == t


@pytest.mark.parametrize(
    "typ,val",
    [
        ("DynArray[DynArray[uint256, 5], 5]", [[], []]),
        ("DynArray[DynArray[DynArray[uint256, 5], 5], 5]", [[[], []], []]),
    ],
)
def test_empty_nested_dynarray(get_contract, typ, val):
    code = f"""
@external
def foo() -> {typ}:
    a: {typ} = {val}
    return a
    """
    c = get_contract(code)
    assert c.foo() == val


# TODO test negative public(DynArray) cases?


# CMC 2022-08-04 these are blocked due to typechecker bug; leaving as
# negative tests so we know if/when the typechecker is fixed.
# (don't consider it a high priority to fix since membership in
# in empty list literal seems like something we should plausibly
# reject at compile-time anyway)
def test_empty_list_membership_fail(get_contract, assert_compile_failed):
    code = """
@external
def foo(x: uint256) -> bool:
    return x in []
    """
    assert_compile_failed(lambda: get_contract(code))
    code = """
@external
def foo(x: uint256) -> bool:
    return x not in []
    """
    assert_compile_failed(lambda: get_contract(code))


# Would be nice to put this somewhere accessible, like in vyper.types or something
integer_types = ["uint8", "int128", "int256", "uint256"]


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant(DynArray[{storage_type}, 3]) = [1, 2, 3]

@external
def foo() -> DynArray[{return_type}, 3]:
    return MY_CONSTANT
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail_2(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant(DynArray[{storage_type}, 3]) = [1, 2, 3]

@external
def foo() -> {return_type}:
    return MY_CONSTANT[0]
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail_3(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant(DynArray[{storage_type}, 3]) = [1, 2, 3]

@external
def foo(i: uint256) -> {return_type}:
    return MY_CONSTANT[i]
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


dynarray_length_no_clobber_cases = [
    # GHSA-3p37-3636-q8wv cases
    """
a: DynArray[uint256,3]

@external
def should_revert() -> DynArray[uint256,3]:
    self.a = [1,2,3]
    self.a = empty(DynArray[uint256,3])
    self.a = [self.a[0], self.a[1], self.a[2]]

    return self.a  # if bug: returns [1,2,3]
    """,
    """
@external
def should_revert() -> DynArray[uint256,3]:
    self.a()
    return self.b() # if bug: returns [1,2,3]

@internal
def a():
    a: uint256 = 0
    b: uint256 = 1
    c: uint256 = 2
    d: uint256 = 3

@internal
def b() -> DynArray[uint256,3]:
    a: DynArray[uint256,3] = empty(DynArray[uint256,3])
    a = [a[0],a[1],a[2]]
    return a
    """,
    """
a: DynArray[uint256,4]

@external
def should_revert() -> DynArray[uint256,4]:
    self.a = [1,2,3]
    self.a = empty(DynArray[uint256,4])
    self.a = [4, self.a[0]]

    return self.a  # if bug: return [4, 4]
    """,
    """
@external
def should_revert() -> DynArray[uint256,4]:
    a: DynArray[uint256, 4] = [1,2,3]
    a = []

    a = [a.pop()]  # if bug: return [1]

    return a
    """,
    """
@external
def should_revert():
    c: DynArray[uint256, 1] = []
    c.append(c[0])
    """,
    """
@external
def should_revert():
    c: DynArray[uint256, 1] = [1]
    c[0] = c.pop()
    """,
    """
@external
def should_revert():
    c: DynArray[DynArray[uint256, 1], 2] = [[]]
    c[0] = c.pop()
    """,
    """
a: DynArray[String[65],2]

@external
def should_revert() -> DynArray[String[65], 2]:
    self.a = ["hello", "world"]
    self.a = []
    self.a = [self.a[0], self.a[1]]

    return self.a  # if bug: return ["hello", "world"]
    """,
]


@pytest.mark.parametrize("code", dynarray_length_no_clobber_cases)
def test_dynarray_length_no_clobber(get_contract, tx_failed, code):
    # check that length is not clobbered before dynarray data copy happens
    c = get_contract(code)
    with tx_failed():
        c.should_revert()



from vyper.utils import OrderedSet
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.analysis.dominators import DominatorTreeAnalysis
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.basicblock import IRBasicBlock, IRInstruction, IROperand, IRVariable
from vyper.venom.passes.base_pass import IRPass


class MakeSSA(IRPass):
    """
    This pass converts the function into Static Single Assignment (SSA) form.
    """

    dom: DominatorTreeAnalysis
    defs: dict[IRVariable, OrderedSet[IRBasicBlock]]

    def run_pass(self):
        fn = self.function

        self.analyses_cache.request_analysis(CFGAnalysis)
        self.dom = self.analyses_cache.request_analysis(DominatorTreeAnalysis)
        self.analyses_cache.request_analysis(LivenessAnalysis)

        self._add_phi_nodes()

        self.var_name_counters = {var.name: 0 for var in self.defs.keys()}
        self.var_name_stacks = {var.name: [0] for var in self.defs.keys()}
        self._rename_vars(fn.entry)
        self._remove_degenerate_phis(fn.entry)

    def _add_phi_nodes(self):
        """
        Add phi nodes to the function.
        """
        self._compute_defs()
        work = {var: 0 for var in self.dom.dfs_walk}
        has_already = {var: 0 for var in self.dom.dfs_walk}
        i = 0

        # Iterate over all variables
        for var, d in self.defs.items():
            i += 1
            defs = list(d)
            while len(defs) > 0:
                bb = defs.pop()
                for dom in self.dom.dominator_frontiers[bb]:
                    if has_already[dom] >= i:
                        continue

                    self._place_phi(var, dom)
                    has_already[dom] = i
                    if work[dom] < i:
                        work[dom] = i
                        defs.append(dom)

    def _place_phi(self, var: IRVariable, basic_block: IRBasicBlock):
        if var not in basic_block.in_vars:
            return

        args: list[IROperand] = []
        for bb in basic_block.cfg_in:
            if bb == basic_block:
                continue

            args.append(bb.label)  # type: ignore
            args.append(var)  # type: ignore

        basic_block.insert_instruction(IRInstruction("phi", args, var), 0)

    def _rename_vars(self, basic_block: IRBasicBlock):
        """
        Rename variables. This follows the placement of phi nodes.
        """
        outs = []

        # Pre-action
        for inst in basic_block.instructions:
            new_ops: list[IROperand] = []
            if inst.opcode != "phi":
                for op in inst.operands:
                    if not isinstance(op, IRVariable):
                        new_ops.append(op)
                        continue

                    new_ops.append(IRVariable(op.name, version=self.var_name_stacks[op.name][-1]))

                inst.operands = new_ops

            if inst.output is not None:
                v_name = inst.output.name
                i = self.var_name_counters[v_name]

                self.var_name_stacks[v_name].append(i)
                self.var_name_counters[v_name] = i + 1

                inst.output = IRVariable(v_name, version=i)
                # note - after previous line, inst.output.name != v_name
                outs.append(inst.output.name)

        for bb in basic_block.cfg_out:
            for inst in bb.instructions:
                if inst.opcode != "phi":
                    continue
                assert inst.output is not None, "Phi instruction without output"
                for i, op in enumerate(inst.operands):
                    if op == basic_block.label:
                        inst.operands[i + 1] = IRVariable(
                            inst.output.name, version=self.var_name_stacks[inst.output.name][-1]
                        )

        for bb in self.dom.dominated[basic_block]:
            if bb == basic_block:
                continue
            self._rename_vars(bb)

        # Post-action
        for op_name in outs:
            # NOTE: each pop corresponds to an append in the pre-action above
            self.var_name_stacks[op_name].pop()

    def _remove_degenerate_phis(self, entry: IRBasicBlock):
        for inst in entry.instructions.copy():
            if inst.opcode != "phi":
                continue

            new_ops: list[IROperand] = []
            for label, op in inst.phi_operands:
                if op == inst.output:
                    continue
                new_ops.extend([label, op])
            new_ops_len = len(new_ops)
            if new_ops_len == 0:
                entry.instructions.remove(inst)
            elif new_ops_len == 2:
                entry.instructions.remove(inst)
            else:
                inst.operands = new_ops

        for bb in self.dom.dominated[entry]:
            if bb == entry:
                continue
            self._remove_degenerate_phis(bb)

    def _compute_defs(self):
        """
        Compute the definition points of variables in the function.
        """
        self.defs = {}
        for bb in self.dom.dfs_walk:
            assignments = bb.get_assignments()
            for var in assignments:
                if var not in self.defs:
                    self.defs[var] = OrderedSet()
                self.defs[var].add(bb)



import re
import textwrap

import hypothesis
import hypothesis.strategies as st
import pytest
from hypothesis import HealthCheck, assume, given
from hypothesis.extra.lark import LarkStrategy

from vyper.ast import Module, parse_to_ast
from vyper.ast.grammar import parse_vyper_source, vyper_grammar
from vyper.ast.pre_parser import pre_parse


def test_basic_grammar():
    code = """
    a: uint256
    b: uint128
    """
    code_func = """
    @external
    def one_two_three() -> uint256:
        return 123123123
    """

    assert parse_vyper_source(code, dedent=True)
    assert parse_to_ast(textwrap.dedent(code))

    assert parse_vyper_source(code_func, dedent=True)
    assert parse_to_ast(textwrap.dedent(code_func))


def test_basic_grammar_empty():
    code = """
    """
    tree = parse_vyper_source(code, dedent=True)
    assert len(tree.children) == 0


def fix_terminal(terminal: str) -> bool:
    # these throw exceptions in the grammar
    for bad in ("\x00", "\\ ", "\x0c"):
        terminal = terminal.replace(bad, " ")
    return terminal


ALLOWED_CHARS = st.characters(codec="ascii", min_codepoint=1)


# With help from hyposmith
# https://github.com/Zac-HD/hypothesmith/blob/master/src/hypothesmith/syntactic.py
class GrammarStrategy(LarkStrategy):
    def __init__(self, grammar, start, explicit_strategies):
        super().__init__(grammar, start, explicit_strategies, alphabet=ALLOWED_CHARS)
        self.terminal_strategies = {
            k: v.map(fix_terminal) for k, v in self.terminal_strategies.items()  # type: ignore
        }

    def draw_symbol(self, data, symbol, draw_state):  # type: ignore
        count = len(draw_state)
        super().draw_symbol(data, symbol, draw_state)
        try:
            compile(
                source="".join(draw_state[count:])
                .replace("contract", "class")
                .replace("struct", "class"),  # HACK: Python ast.parse
                filename="<string>",
                mode="exec",
            )
        except SyntaxError:
            # Python's grammar doesn't actually fully describe the behaviour of the
            # CPython parser and AST-post-processor, so we just filter out errors.
            assume(False)


def from_grammar() -> st.SearchStrategy[str]:
    """
    Generate syntactically-valid Python source code based on the grammar.
    """
    grammar = vyper_grammar()
    explicit_strategies = dict(
        _INDENT=st.just(" " * 4),
        _DEDENT=st.just(""),
        NAME=st.from_regex(r"[a-z_A-Z]+", fullmatch=True).filter(str.isidentifier),
    )
    return GrammarStrategy(grammar, "module", explicit_strategies)


# Avoid examples with *only* single or double quote docstrings
# because they trigger a trivial parser bug
SINGLE_QUOTE_DOCSTRING = re.compile(r"^'''.*'''$")
DOUBLE_QUOTE_DOCSTRING = re.compile(r'^""".*"""$')


def has_no_docstrings(c):
    return not (SINGLE_QUOTE_DOCSTRING.match(c) or DOUBLE_QUOTE_DOCSTRING.match(c))


@pytest.mark.fuzzing
@given(code=from_grammar())
@hypothesis.settings(
    max_examples=500, suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much]
)
def test_grammar_bruteforce(code):
    _, _, _, reformatted_code = pre_parse(code + "\n")
    tree = parse_to_ast(reformatted_code)
    assert isinstance(tree, Module)



import contextlib
import json
import posixpath
from dataclasses import asdict, dataclass, field
from functools import cached_property
from pathlib import Path, PurePath
from typing import TYPE_CHECKING, Any, Iterator, Optional

from vyper.exceptions import JSONError
from vyper.utils import sha256sum

# a type to make mypy happy
PathLike = Path | PurePath

if TYPE_CHECKING:
    from zipfile import ZipFile


@dataclass(frozen=True)
class CompilerInput:
    # an input to the compiler, basically an abstraction for file contents

    source_id: int
    path: PathLike  # the path that was asked for

    # resolved_path is the real path that was resolved to.
    # mainly handy for debugging at this point
    resolved_path: PathLike
    contents: str

    @cached_property
    def sha256sum(self):
        return sha256sum(self.contents)


@dataclass(frozen=True)
class FileInput(CompilerInput):
    @cached_property
    def source_code(self):
        return self.contents


@dataclass(frozen=True, unsafe_hash=True)
class ABIInput(CompilerInput):
    # some json input, which has already been parsed into a dict or list
    # this is needed because json inputs present json interfaces as json
    # objects, not as strings. this class helps us avoid round-tripping
    # back to a string to pretend it's a file.
    abi: Any = field(hash=False)  # something that json.load() returns


def try_parse_abi(file_input: FileInput) -> CompilerInput:
    try:
        s = json.loads(file_input.source_code)
        return ABIInput(**asdict(file_input), abi=s)
    except (ValueError, TypeError):
        return file_input


class _NotFound(Exception):
    pass


# an opaque object which consumers can get/set attributes on
class _Cache(object):
    pass


# an "input bundle" to the compiler, representing the files which are
# available to the compiler. it is useful because it parametrizes I/O
# operations over different possible input types. you can think of it
# as a virtual filesystem which models the compiler's interactions
# with the outside world. it exposes a "load_file" operation which
# searches for a file from a set of search paths, and also provides
# id generation service to get a unique source id per file.
class InputBundle:
    # a list of search paths
    search_paths: list[PathLike]

    _cache: Any

    def __init__(self, search_paths):
        self.search_paths = search_paths
        self._source_id_counter = 0
        self._source_ids: dict[PathLike, int] = {}

        # this is a little bit cursed, but it allows consumers to cache data
        # that share the same lifetime as this input bundle.
        self._cache = _Cache()

    def _normalize_path(self, path):
        raise NotImplementedError(f"not implemented! {self.__class__}._normalize_path()")

    def _load_from_path(self, resolved_path, path):
        raise NotImplementedError(f"not implemented! {self.__class__}._load_from_path()")

    def _generate_source_id(self, resolved_path: PathLike) -> int:
        # Note: it is possible for a file to get in here more than once,
        # e.g. by symlink
        if resolved_path not in self._source_ids:
            self._source_ids[resolved_path] = self._source_id_counter
            self._source_id_counter += 1

        return self._source_ids[resolved_path]

    def load_file(self, path: PathLike | str) -> CompilerInput:
        # search path precedence
        tried = []
        if isinstance(path, str):
            path = PurePath(path)
        for sp in reversed(self.search_paths):
            # note from pathlib docs:
            # > If the argument is an absolute path, the previous path is ignored.
            # Path("/a") / Path("/b") => Path("/b")
            to_try = sp / path

            try:
                to_try = self._normalize_path(to_try)
                res = self._load_from_path(to_try, path)
                break
            except _NotFound:
                tried.append(to_try)

        else:
            formatted_search_paths = "\n".join(["  " + str(p) for p in tried])
            raise FileNotFoundError(
                f"could not find {path} in any of the following locations:\n"
                f"{formatted_search_paths}"
            )

        # try to parse from json, so that return types are consistent
        # across FilesystemInputBundle and JSONInputBundle.
        if isinstance(res, FileInput):
            res = try_parse_abi(res)

        return res

    def add_search_path(self, path: PathLike) -> None:
        self.search_paths.append(path)

    # temporarily add something to the search path (within the
    # scope of the context manager) with highest precedence.
    # if `path` is None, do nothing
    @contextlib.contextmanager
    def search_path(self, path: Optional[PathLike]) -> Iterator[None]:
        if path is None:
            yield  # convenience, so caller does not have to handle null path

        else:
            self.search_paths.append(path)
            try:
                yield
            finally:
                self.search_paths.pop()

    # temporarily modify the top of the search path (within the
    # scope of the context manager) with highest precedence to something else
    @contextlib.contextmanager
    def poke_search_path(self, path: PathLike) -> Iterator[None]:
        tmp = self.search_paths[-1]
        self.search_paths[-1] = path
        try:
            yield
        finally:
            self.search_paths[-1] = tmp


# regular input. takes a search path(s), and `load_file()` will search all
# search paths for the file and read it from the filesystem
class FilesystemInputBundle(InputBundle):
    def _normalize_path(self, path: Path) -> Path:
        # normalize the path with os.path.normpath, to break down
        # things like "foo/bar/../x.vy" => "foo/x.vy", with all
        # the caveats around symlinks that os.path.normpath comes with.
        try:
            return path.resolve(strict=True)
        except (FileNotFoundError, NotADirectoryError):
            raise _NotFound(path)

    def _load_from_path(self, resolved_path: Path, original_path: Path) -> CompilerInput:
        try:
            with resolved_path.open() as f:
                code = f.read()
        except (FileNotFoundError, NotADirectoryError):
            raise _NotFound(resolved_path)

        source_id = super()._generate_source_id(resolved_path)

        return FileInput(source_id, original_path, resolved_path, code)


# wrap os.path.normpath, but return the same type as the input -
# but use posixpath instead so that things work cross-platform.
def _normpath(path):
    cls = path.__class__
    if not isinstance(path, str):
        path = path.as_posix()
    return cls(posixpath.normpath(path))


# fake filesystem for "standard JSON" (aka solc-style) inputs. takes search
# paths, and `load_file()` "reads" the file from the JSON input. Note that this
# input bundle type never actually interacts with the filesystem -- it is
# guaranteed to be pure!
class JSONInputBundle(InputBundle):
    input_json: dict[PurePath, Any]

    def __init__(self, input_json, search_paths):
        super().__init__(search_paths)
        self.input_json = {}
        for path, item in input_json.items():
            path = _normpath(path)

            # should be checked by caller
            assert path not in self.input_json
            self.input_json[path] = item

    def _normalize_path(self, path: PurePath) -> PurePath:
        return _normpath(path)

    def _load_from_path(self, resolved_path: PurePath, original_path: PurePath) -> CompilerInput:
        try:
            value = self.input_json[resolved_path]
        except KeyError:
            raise _NotFound(resolved_path)

        source_id = super()._generate_source_id(resolved_path)

        if "content" in value:
            return FileInput(source_id, original_path, resolved_path, value["content"])

        if "abi" in value:
            return ABIInput(
                source_id, original_path, resolved_path, json.dumps(value), value["abi"]
            )

        # TODO: ethPM support
        # if isinstance(contents, dict) and "contractTypes" in contents:

        # unreachable, based on how JSONInputBundle is constructed in
        # the codebase.
        raise JSONError(f"Unexpected type in file: '{resolved_path}'")  # pragma: nocover


# input bundle for vyper archives. similar to JSONInputBundle, but takes
# a zipfile as input.
class ZipInputBundle(InputBundle):
    def __init__(self, archive: "ZipFile"):
        assert archive.testzip() is None
        self.archive = archive

        sp_str = archive.read("MANIFEST/searchpaths").decode("utf-8")
        search_paths = [PurePath(p) for p in sp_str.splitlines()]

        super().__init__(search_paths)

    def _normalize_path(self, path: PurePath) -> PurePath:
        return _normpath(path)

    def _load_from_path(self, resolved_path: PurePath, original_path: PurePath) -> CompilerInput:
        # zipfile.BadZipFile: File is not a zip file

        try:
            value = self.archive.read(resolved_path.as_posix()).decode("utf-8")
        except KeyError:
            # zipfile literally raises KeyError if the file is not there
            raise _NotFound(resolved_path)

        source_id = super()._generate_source_id(resolved_path)

        return FileInput(source_id, original_path, resolved_path, value)



import pytest

from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import version_check
from vyper.ir import compile_ir
from vyper.ir.s_expressions import parse_s_exp

fail_list = [
    [-(2**255) - 3],
    [2**256 + 3],
    ["set", "_poz"],
    [["set", "var_1", 0, 0]],
    ["with", "var_1", 0, ["set", 1, 1]],
    ["break"],  # invalid break
    ["continue"],  # invalid continue
    ["invalidllelement"],
]


@pytest.mark.parametrize("bad_ir", fail_list)
def test_ir_compile_fail(bad_ir, get_contract_from_ir, assert_compile_failed):
    assert_compile_failed(lambda: get_contract_from_ir(IRnode.from_list(bad_ir)), Exception)


valid_list = [
    ["pass"],
    ["assert", ["slt", ["mload", 0], 300]],
    ["assert", ["sgt", ["mload", 0], -1]],
    ["assert", ["gt", 1, ["mload", 0]]],
    ["assert", ["ge", ["mload", 0], 0]],
]


@pytest.mark.parametrize("good_ir", valid_list)
def test_compile_ir_good(good_ir, get_contract_from_ir):
    get_contract_from_ir(IRnode.from_list(good_ir))


def test_ir_from_s_expression(get_contract_from_ir):
    code = """
(seq
  (deploy
    0
    (seq ; just return 32 byte of calldata back
      (calldatacopy 0 4 32)
      (return 0 32)
      stop
     )
    0))
    """
    abi = [
        {
            "name": "test",
            "outputs": [{"type": "int128", "name": "out"}],
            "inputs": [{"type": "int128", "name": "a"}],
            "stateMutability": "nonpayable",
            "type": "function",
            "gas": 394,
        }
    ]

    s_expressions = parse_s_exp(code)
    ir = IRnode.from_list(s_expressions[0])
    c = get_contract_from_ir(ir, abi=abi)
    assert c.test(-123456) == -123456


def test_pc_debugger():
    debugger_ir = ["seq", ["mstore", 0, 32], ["pc_debugger"]]
    ir_nodes = IRnode.from_list(debugger_ir)
    _, line_number_map = compile_ir.assembly_to_evm(compile_ir.compile_to_assembly(ir_nodes))
    if version_check(begin="shanghai"):
        offset = 4  # push0 saves a byte
    else:
        offset = 5

    assert line_number_map["pc_breakpoints"][0] == offset



import enum

from vyper.utils import StringEnum


class DataLocation(StringEnum):
    UNSET = enum.auto()
    MEMORY = enum.auto()
    STORAGE = enum.auto()
    CALLDATA = enum.auto()
    CODE = enum.auto()
    TRANSIENT = enum.auto()



import pytest

AMOUNT = 100 * 10**18
TOKEN_NAME = "Vypercoin"
TOKEN_SYMBOL = "FANG"
TOKEN_DECIMALS = 18
TOKEN_INITIAL_SUPPLY = 0


@pytest.fixture(scope="module")
def token(get_contract):
    with open("examples/tokens/ERC20.vy") as f:
        return get_contract(
            f.read(), TOKEN_NAME, TOKEN_SYMBOL, TOKEN_DECIMALS, TOKEN_INITIAL_SUPPLY
        )


@pytest.fixture(scope="module")
def vault(get_contract, token):
    with open("examples/tokens/ERC4626.vy") as f:
        return get_contract(f.read(), token.address)


def test_asset(vault, token):
    assert vault.asset() == token.address


def test_max_methods(env, vault):
    a = env.accounts[0]

    assert vault.maxDeposit(a) == 2**256 - 1
    assert vault.maxMint(a) == 2**256 - 1
    assert vault.maxWithdraw(a) == 2**256 - 1
    assert vault.maxRedeem(a) == 2**256 - 1


def test_preview_methods(env, token, vault):
    a = env.accounts[0]

    assert vault.totalAssets() == 0
    assert vault.convertToAssets(10**18) == 0  # no assets
    assert vault.convertToShares(10**18) == 10**18  # 1:1 price
    assert vault.previewDeposit(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewMint(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewWithdraw(AMOUNT) == 0  # but no assets
    assert vault.previewRedeem(AMOUNT) == 0  # but no assets

    token.mint(a, AMOUNT, sender=a)
    token.approve(vault.address, AMOUNT, sender=a)
    vault.deposit(AMOUNT, sender=a)

    assert vault.totalAssets() == AMOUNT
    assert vault.convertToAssets(10**18) == 10**18  # 1:1 price
    assert vault.convertToShares(10**18) == 10**18  # 1:1 price
    assert vault.previewDeposit(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewMint(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewWithdraw(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewRedeem(AMOUNT) == AMOUNT  # 1:1 price

    token.mint(vault.address, AMOUNT, sender=a)

    assert vault.totalAssets() == 2 * AMOUNT
    assert vault.convertToAssets(10**18) == 2 * 10**18  # 2:1 price
    assert vault.convertToShares(2 * 10**18) == 10**18  # 2:1 price
    assert vault.previewDeposit(AMOUNT) == AMOUNT // 2  # 2:1 price
    assert vault.previewMint(AMOUNT // 2) == AMOUNT  # 2:1 price
    assert vault.previewWithdraw(AMOUNT) == AMOUNT // 2  # 2:1 price
    assert vault.previewRedeem(AMOUNT // 2) == AMOUNT  # 2:1 price

    vault.DEBUG_steal_tokens(AMOUNT, sender=a)

    assert vault.totalAssets() == AMOUNT
    assert vault.convertToAssets(10**18) == 10**18  # 1:1 price
    assert vault.convertToShares(10**18) == 10**18  # 1:1 price
    assert vault.previewDeposit(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewMint(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewWithdraw(AMOUNT) == AMOUNT  # 1:1 price
    assert vault.previewRedeem(AMOUNT) == AMOUNT  # 1:1 price

    vault.DEBUG_steal_tokens(AMOUNT // 2, sender=a)

    assert vault.totalAssets() == AMOUNT // 2
    assert vault.convertToAssets(10**18) == 10**18 // 2  # 1:2 price
    assert vault.convertToShares(10**18 // 2) == 10**18  # 1:2 price
    assert vault.previewDeposit(AMOUNT) == 2 * AMOUNT  # 1:2 price
    assert vault.previewMint(2 * AMOUNT) == AMOUNT  # 1:2 price
    assert vault.previewWithdraw(AMOUNT) == 2 * AMOUNT  # 1:2 price
    assert vault.previewRedeem(2 * AMOUNT) == AMOUNT  # 1:2 price



from vyper.semantics.types.primitives import SINT, UINT, BytesM_T, IntegerT

# shortcut type names
UINT256_T = IntegerT(False, 256)
UINT8_T = IntegerT(False, 8)
INT256_T = IntegerT(True, 256)
INT128_T = IntegerT(True, 128)
UINT160_T = IntegerT(False, 160)

BYTES32_T = BytesM_T(32)
BYTES20_T = BytesM_T(20)
BYTES4_T = BytesM_T(4)

_ = UINT, SINT  # explicitly use: addresses linter F401



def test_keccak_sanity(keccak):
    # sanity check -- ensure keccak is keccak256, not sha3
    # https://ethereum.stackexchange.com/a/107985
    assert keccak(b"").hex() == "c5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"



# CMC 2024-02-03 TODO: rename me to function.py

import contextlib
from typing import Optional

from vyper import ast as vy_ast
from vyper.ast.validation import validate_call_args
from vyper.exceptions import (
    CallViolation,
    ExceptionList,
    FunctionDeclarationException,
    ImmutableViolation,
    InvalidType,
    IteratorException,
    NonPayableViolation,
    StateAccessViolation,
    StructureException,
    TypeCheckFailure,
    TypeMismatch,
    VariableDeclarationException,
    VyperException,
)

# TODO consolidate some of these imports
from vyper.semantics.analysis.base import (
    Modifiability,
    ModuleInfo,
    ModuleOwnership,
    VarAccess,
    VarInfo,
)
from vyper.semantics.analysis.common import VyperNodeVisitorBase
from vyper.semantics.analysis.utils import (
    get_common_types,
    get_exact_type_from_node,
    get_expr_info,
    get_possible_types_from_node,
    uses_state,
    validate_expected_type,
)
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.environment import CONSTANT_ENVIRONMENT_VARS
from vyper.semantics.namespace import get_namespace
from vyper.semantics.types import (
    TYPE_T,
    VOID_TYPE,
    AddressT,
    BoolT,
    DArrayT,
    EventT,
    FlagT,
    HashMapT,
    IntegerT,
    SArrayT,
    SelfT,
    StringT,
    StructT,
    TupleT,
    VyperType,
    _BytestringT,
    is_type_t,
    map_void,
)
from vyper.semantics.types.function import ContractFunctionT, MemberFunctionT, StateMutability
from vyper.semantics.types.utils import type_from_annotation


def analyze_functions(vy_module: vy_ast.Module) -> None:
    """Analyzes a vyper ast and validates the function bodies"""
    err_list = ExceptionList()

    for node in vy_module.get_children(vy_ast.FunctionDef):
        _analyze_function_r(vy_module, node, err_list)

    for node in vy_module.get_children(vy_ast.VariableDecl):
        if not node.is_public:
            continue
        _analyze_function_r(vy_module, node._expanded_getter, err_list)

    err_list.raise_if_not_empty()


def _analyze_function_r(
    vy_module: vy_ast.Module, node: vy_ast.FunctionDef, err_list: ExceptionList
):
    func_t = node._metadata["func_type"]

    for call_t in func_t.called_functions:
        if isinstance(call_t, ContractFunctionT):
            assert isinstance(call_t.ast_def, vy_ast.FunctionDef)  # help mypy
            _analyze_function_r(vy_module, call_t.ast_def, err_list)

    namespace = get_namespace()

    try:
        with namespace.enter_scope():
            analyzer = FunctionAnalyzer(vy_module, node, namespace)
            analyzer.analyze()
    except VyperException as e:
        err_list.append(e)


# finds the terminus node for a list of nodes.
# raises an exception if any nodes are unreachable
def find_terminating_node(node_list: list) -> Optional[vy_ast.VyperNode]:
    ret = None

    for node in node_list:
        if ret is not None:
            raise StructureException("Unreachable code!", node)

        if node.is_terminus:
            ret = node

        if isinstance(node, vy_ast.If):
            body_terminates = find_terminating_node(node.body)

            else_terminates = None
            if node.orelse is not None:
                else_terminates = find_terminating_node(node.orelse)

            if body_terminates is not None and else_terminates is not None:
                ret = else_terminates

        if isinstance(node, vy_ast.For):
            # call find_terminating_node for its side effects
            find_terminating_node(node.body)

    return ret


# helpers
def _validate_address_code(node: vy_ast.Attribute, value_type: VyperType) -> None:
    if isinstance(value_type, AddressT) and node.attr == "code":
        # Validate `slice(<address>.code, start, length)` where `length` is constant
        parent = node.get_ancestor()
        if isinstance(parent, vy_ast.Call):
            ok_func = isinstance(parent.func, vy_ast.Name) and parent.func.id == "slice"
            ok_args = len(parent.args) == 3 and isinstance(parent.args[2], vy_ast.Int)
            if ok_func and ok_args:
                return

        raise StructureException(
            "(address).code is only allowed inside of a slice function with a constant length", node
        )


def _validate_msg_data_attribute(node: vy_ast.Attribute) -> None:
    if isinstance(node.value, vy_ast.Name) and node.value.id == "msg" and node.attr == "data":
        parent = node.get_ancestor()
        allowed_builtins = ("slice", "len", "raw_call")
        if not isinstance(parent, vy_ast.Call) or parent.get("func.id") not in allowed_builtins:
            raise StructureException(
                "msg.data is only allowed inside of the slice, len or raw_call functions", node
            )
        if parent.get("func.id") == "slice":
            ok_args = len(parent.args) == 3 and isinstance(parent.args[2], vy_ast.Int)
            if not ok_args:
                raise StructureException(
                    "slice(msg.data) must use a compile-time constant for length argument", parent
                )


def _validate_msg_value_access(node: vy_ast.Attribute) -> None:
    if isinstance(node.value, vy_ast.Name) and node.attr == "value" and node.value.id == "msg":
        raise NonPayableViolation("msg.value is not allowed in non-payable functions", node)


def _validate_pure_access(node: vy_ast.Attribute | vy_ast.Name, typ: VyperType) -> None:
    if isinstance(typ, TYPE_T):
        return

    info = get_expr_info(node)

    env_vars = CONSTANT_ENVIRONMENT_VARS
    # check env variable access like `block.number`
    if isinstance(node, vy_ast.Attribute):
        if node.get("value.id") in env_vars:
            raise StateAccessViolation(
                "not allowed to query environment variables in pure functions"
            )
        parent_info = get_expr_info(node.value)
        if isinstance(parent_info.typ, AddressT) and node.attr in AddressT._type_members:
            raise StateAccessViolation("not allowed to query address members in pure functions")

    if (varinfo := info.var_info) is None:
        return
    # self is magic. we only need to check it if it is not the root of an Attribute
    # node. (i.e. it is bare like `self`, not `self.foo`)
    is_naked_self = isinstance(varinfo.typ, SelfT) and not isinstance(
        node.get_ancestor(), vy_ast.Attribute
    )
    if is_naked_self:
        raise StateAccessViolation("not allowed to query `self` in pure functions")

    if varinfo.is_state_variable() or is_naked_self:
        raise StateAccessViolation("not allowed to query state variables in pure functions")


# analyse the variable access for the attribute chain for a node
# e.x. `x` will return varinfo for `x`
# `module.foo` will return VarAccess for `module.foo`
# `self.my_struct.x.y` will return VarAccess for `self.my_struct.x.y`
def _get_variable_access(node: vy_ast.ExprNode) -> Optional[VarAccess]:
    path: list[str | object] = []
    info = get_expr_info(node)

    while info.var_info is None:
        if not isinstance(node, (vy_ast.Subscript, vy_ast.Attribute)):
            # it's something like a literal
            return None

        if isinstance(node, vy_ast.Subscript):
            # Subscript is an analysis barrier
            # we cannot analyse if `x.y[ix1].z` overlaps with `x.y[ix2].z`.
            path.append(VarAccess.SUBSCRIPT_ACCESS)

        if (attr := info.attr) is not None:
            path.append(attr)

        assert isinstance(node, (vy_ast.Subscript, vy_ast.Attribute))  # help mypy
        node = node.value
        info = get_expr_info(node)

    # ignore `self.` as it interferes with VarAccess comparison across modules
    if len(path) > 0 and path[-1] == "self":
        path.pop()
    path.reverse()

    return VarAccess(info.var_info, tuple(path))


# get the chain of modules, e.g.
# mod1.mod2.x.y -> [ModuleInfo(mod1), ModuleInfo(mod2)]
# CMC 2024-02-12 note that the Attribute/Subscript traversal in this and
# _get_variable_access() are a bit gross and could probably
# be refactored into data on ExprInfo.
def _get_module_chain(node: vy_ast.ExprNode) -> list[ModuleInfo]:
    ret: list[ModuleInfo] = []
    info = get_expr_info(node)

    while True:
        if info.module_info is not None:
            ret.append(info.module_info)

        if not isinstance(node, (vy_ast.Subscript, vy_ast.Attribute)):
            break

        node = node.value
        info = get_expr_info(node)

    ret.reverse()
    return ret


def check_module_uses(node: vy_ast.ExprNode) -> Optional[ModuleInfo]:
    """
    validate module usage, and that if we use lib1.lib2.<state>, that
    lib1 at least `uses` lib2.

    Returns the left-most module referenced in the expr,
        e.g. `lib1.lib2.foo` should return module info for `lib1`.
    """
    module_infos = _get_module_chain(node)

    if len(module_infos) == 0:
        return None

    for module_info in module_infos:
        if module_info.ownership < ModuleOwnership.USES:
            msg = f"Cannot access `{module_info.alias}` state!\n  note that"
            # CMC 2024-04-12 add UX note about nonreentrant. might be nice
            # in the future to be more specific about exactly which state is
            # used, although that requires threading a bit more context into
            # this function.
            msg += " use of the `@nonreentrant` decorator is also considered"
            msg += " state access"

            hint = f"add `uses: {module_info.alias}` or "
            hint += f"`initializes: {module_info.alias}` as "
            hint += "a top-level statement to your contract"
            raise ImmutableViolation(msg, hint=hint)

    # the leftmost- referenced module
    root_module_info = module_infos[0]
    return root_module_info


class FunctionAnalyzer(VyperNodeVisitorBase):
    ignored_types = (vy_ast.Pass,)
    scope_name = "function"

    def __init__(
        self, vyper_module: vy_ast.Module, fn_node: vy_ast.FunctionDef, namespace: dict
    ) -> None:
        self.vyper_module = vyper_module
        self.fn_node = fn_node
        self.namespace = namespace
        self.func = fn_node._metadata["func_type"]
        self.expr_visitor = ExprVisitor(self)

        self.loop_variables: list[Optional[VarAccess]] = []

    def analyze(self):
        if self.func.analysed:
            return

        # mark seen before analysing, if analysis throws an exception which
        # gets caught, we don't want to analyse again.
        self.func.mark_analysed()

        # allow internal function params to be mutable
        if self.func.is_internal:
            location, modifiability = (DataLocation.MEMORY, Modifiability.MODIFIABLE)
        else:
            location, modifiability = (DataLocation.CALLDATA, Modifiability.RUNTIME_CONSTANT)

        for arg in self.func.arguments:
            self.namespace[arg.name] = VarInfo(
                arg.typ, location=location, modifiability=modifiability
            )

        for node in self.fn_node.body:
            self.visit(node)

        if self.func.return_type:
            if not find_terminating_node(self.fn_node.body):
                raise FunctionDeclarationException(
                    f"Missing return statement in function '{self.fn_node.name}'", self.fn_node
                )
        else:
            # call find_terminator for its unreachable code detection side effect
            find_terminating_node(self.fn_node.body)

        # visit default args
        assert self.func.n_keyword_args == len(self.fn_node.args.defaults)
        for kwarg in self.func.keyword_args:
            self.expr_visitor.visit(kwarg.default_value, kwarg.typ)

    @contextlib.contextmanager
    def enter_for_loop(self, varaccess: Optional[VarAccess]):
        self.loop_variables.append(varaccess)
        try:
            yield
        finally:
            self.loop_variables.pop()

    def visit(self, node):
        super().visit(node)

    def visit_AnnAssign(self, node):
        name = node.get("target.id")
        if name is None:
            raise VariableDeclarationException("Invalid assignment", node)

        if not node.value:
            raise VariableDeclarationException(
                "Memory variables must be declared with an initial value", node
            )

        typ = type_from_annotation(node.annotation, DataLocation.MEMORY)

        # validate the value before adding it to the namespace
        self.expr_visitor.visit(node.value, typ)

        self.namespace[name] = VarInfo(typ, location=DataLocation.MEMORY)

        self.expr_visitor.visit(node.target, typ)

    def _validate_revert_reason(self, msg_node: vy_ast.VyperNode) -> None:
        if isinstance(msg_node, vy_ast.Str):
            if not msg_node.value.strip():
                raise StructureException("Reason string cannot be empty", msg_node)
            self.expr_visitor.visit(msg_node, get_exact_type_from_node(msg_node))
        elif not (isinstance(msg_node, vy_ast.Name) and msg_node.id == "UNREACHABLE"):
            try:
                validate_expected_type(msg_node, StringT(1024))
            except TypeMismatch as e:
                raise InvalidType("revert reason must fit within String[1024]") from e
            self.expr_visitor.visit(msg_node, get_exact_type_from_node(msg_node))
        # CMC 2023-10-19 nice to have: tag UNREACHABLE nodes with a special type

    def visit_Assert(self, node):
        if node.msg:
            self._validate_revert_reason(node.msg)

        self.expr_visitor.visit(node.test, BoolT())

    # repeated code for Assign and AugAssign
    def _assign_helper(self, node):
        if isinstance(node.value, vy_ast.Tuple):
            raise StructureException("Right-hand side of assignment cannot be a tuple", node.value)

        target = get_expr_info(node.target)

        # check mutability of the function
        self._handle_modification(node.target)

        self.expr_visitor.visit(node.value, target.typ)
        self.expr_visitor.visit(node.target, target.typ)

    def _handle_modification(self, target: vy_ast.ExprNode):
        if isinstance(target, vy_ast.Tuple):
            for item in target.elements:
                self._handle_modification(item)
            return

        # check a modification of `target`. validate the modification is
        # valid, and log the modification in relevant data structures.
        func_t = self.func
        info = get_expr_info(target)

        if isinstance(info.typ, HashMapT):
            raise StructureException(
                "Left-hand side of assignment cannot be a HashMap without a key"
            )

        if (
            info.location in (DataLocation.STORAGE, DataLocation.TRANSIENT)
            and func_t.mutability <= StateMutability.VIEW
        ):
            raise StateAccessViolation(
                f"Cannot modify {info.location} variable in a {func_t.mutability} function"
            )

        if info.location == DataLocation.CALLDATA:
            raise ImmutableViolation("Cannot write to calldata")

        if info.modifiability == Modifiability.RUNTIME_CONSTANT:
            if info.location == DataLocation.CODE:
                if not func_t.is_constructor:
                    raise ImmutableViolation("Immutable value cannot be written to")

                # handle immutables
                if info.var_info is not None:  # don't handle complex (struct,array) immutables
                    # special handling for immutable variables in the ctor
                    # TODO: maybe we want to remove this restriction.
                    if info.var_info._modification_count != 0:
                        raise ImmutableViolation(
                            "Immutable value cannot be modified after assignment"
                        )
                    info.var_info._modification_count += 1
            else:
                raise ImmutableViolation("Environment variable cannot be written to")

        if info.modifiability == Modifiability.CONSTANT:
            raise ImmutableViolation("Constant value cannot be written to.")

        var_access = _get_variable_access(target)
        assert var_access is not None

        info._writes.add(var_access)

    def _handle_module_access(self, target: vy_ast.ExprNode):
        root_module_info = check_module_uses(target)

        if root_module_info is not None:
            # log the access
            self.func.mark_used_module(root_module_info)

    def visit_Assign(self, node):
        self._assign_helper(node)

    def visit_AugAssign(self, node):
        self._assign_helper(node)
        node.target._expr_info.typ.validate_numeric_op(node)

    def visit_Break(self, node):
        for_node = node.get_ancestor(vy_ast.For)
        if for_node is None:
            raise StructureException("`break` must be enclosed in a `for` loop", node)

    def visit_Continue(self, node):
        # TODO: use context/state instead of ast search
        for_node = node.get_ancestor(vy_ast.For)
        if for_node is None:
            raise StructureException("`continue` must be enclosed in a `for` loop", node)

    def visit_Expr(self, node):
        if isinstance(node.value, vy_ast.Ellipsis):
            raise StructureException(
                "`...` is not allowed in `.vy` files! "
                "Did you mean to import me as a `.vyi` file?",
                node,
            )

        # NOTE: standalone staticcalls are banned!
        if not isinstance(node.value, (vy_ast.Call, vy_ast.ExtCall)):
            raise StructureException(
                "Expressions without assignment are disallowed",
                node,
                hint="did you mean to assign the result to a variable?",
            )

        if isinstance(node.value, vy_ast.ExtCall):
            call_node = node.value.value
        else:
            call_node = node.value

        func = call_node.func

        fn_type = get_exact_type_from_node(func)

        if is_type_t(fn_type, EventT):
            raise StructureException("To call an event you must use the `log` statement", node)

        if is_type_t(fn_type, StructT):
            raise StructureException("Struct creation without assignment is disallowed", node)

        # NOTE: fetch_call_return validates call args.
        return_value = map_void(fn_type.fetch_call_return(call_node))
        if (
            return_value is not VOID_TYPE
            and not isinstance(fn_type, MemberFunctionT)
            and not isinstance(fn_type, ContractFunctionT)
        ):
            raise StructureException(
                f"Function `{fn_type}` cannot be called without assigning the result"
            )
        self.expr_visitor.visit(node.value, return_value)

    def _analyse_range_iter(self, iter_node, target_type):
        # iteration via range()
        if iter_node.get("func.id") != "range":
            raise IteratorException("Cannot iterate over the result of a function call", iter_node)
        _validate_range_call(iter_node)

        args = iter_node.args
        kwargs = [s.value for s in iter_node.keywords]
        for arg in (*args, *kwargs):
            self.expr_visitor.visit(arg, target_type)

    def _analyse_list_iter(self, iter_node, target_type):
        # iteration over a variable or literal list
        iter_val = iter_node.reduced()

        if isinstance(iter_val, vy_ast.List):
            len_ = len(iter_val.elements)
            if len_ == 0:
                raise StructureException("For loop must have at least 1 iteration", iter_node)
            iter_type = SArrayT(target_type, len_)
        else:
            try:
                iter_type = get_exact_type_from_node(iter_node)
            except (InvalidType, StructureException):
                raise InvalidType("Not an iterable type", iter_node)

        # CMC 2024-02-09 TODO: use validate_expected_type once we have DArrays
        # with generic length.
        if not isinstance(iter_type, (DArrayT, SArrayT)):
            raise InvalidType("Not an iterable type", iter_node)

        self.expr_visitor.visit(iter_node, iter_type)

        # get the root varinfo from iter_val in case we need to peer
        # through folded constants
        return _get_variable_access(iter_val)

    def visit_For(self, node):
        if not isinstance(node.target.target, vy_ast.Name):
            raise StructureException("Invalid syntax for loop iterator", node.target.target)

        target_type = type_from_annotation(node.target.annotation, DataLocation.MEMORY)

        iter_var = None
        if isinstance(node.iter, vy_ast.Call):
            self._analyse_range_iter(node.iter, target_type)

            # sanity check the postcondition of analyse_range_iter
            assert isinstance(target_type, IntegerT)
        else:
            iter_var = self._analyse_list_iter(node.iter, target_type)

        with self.namespace.enter_scope(), self.enter_for_loop(iter_var):
            target_name = node.target.target.id
            # maybe we should introduce a new Modifiability: LOOP_VARIABLE
            self.namespace[target_name] = VarInfo(
                target_type, modifiability=Modifiability.RUNTIME_CONSTANT
            )

            self.expr_visitor.visit(node.target.target, target_type)

            for stmt in node.body:
                self.visit(stmt)

    def visit_If(self, node):
        self.expr_visitor.visit(node.test, BoolT())
        with self.namespace.enter_scope():
            for n in node.body:
                self.visit(n)
        with self.namespace.enter_scope():
            for n in node.orelse:
                self.visit(n)

    def visit_Log(self, node):
        # postcondition of Log.validate()
        assert isinstance(node.value, vy_ast.Call)

        f = get_exact_type_from_node(node.value.func)
        if not is_type_t(f, EventT):
            raise StructureException("Value is not an event", node.value)
        if self.func.mutability <= StateMutability.VIEW:
            raise StructureException(
                f"Cannot emit logs from {self.func.mutability} functions", node
            )
        t = map_void(f.fetch_call_return(node.value))
        # CMC 2024-02-05 annotate the event type for codegen usage
        # TODO: refactor this
        node._metadata["type"] = f.typedef
        self.expr_visitor.visit(node.value, t)

    def visit_Raise(self, node):
        if node.exc:
            self._validate_revert_reason(node.exc)

    def visit_Return(self, node):
        values = node.value
        if values is None:
            if self.func.return_type:
                raise FunctionDeclarationException("Return statement is missing a value", node)
            return
        elif self.func.return_type is None:
            raise FunctionDeclarationException("Function should not return any values", node)

        if isinstance(values, vy_ast.Tuple):
            values = values.elements
            if not isinstance(self.func.return_type, TupleT):
                raise FunctionDeclarationException("Function only returns a single value", node)
            if self.func.return_type.length != len(values):
                raise FunctionDeclarationException(
                    f"Incorrect number of return values: "
                    f"expected {self.func.return_type.length}, got {len(values)}",
                    node,
                )

        self.expr_visitor.visit(node.value, self.func.return_type)


class ExprVisitor(VyperNodeVisitorBase):
    def __init__(self, function_analyzer: Optional[FunctionAnalyzer] = None):
        self.function_analyzer = function_analyzer

    @property
    def func(self):
        if self.function_analyzer is None:
            return None
        return self.function_analyzer.func

    @property
    def scope_name(self):
        if self.func is not None:
            return "function"
        return "module"

    def visit(self, node, typ):
        if typ is not VOID_TYPE and not isinstance(typ, TYPE_T):
            validate_expected_type(node, typ)

        # recurse and typecheck in case we are being fed the wrong type for
        # some reason.
        super().visit(node, typ)

        # annotate
        node._metadata["type"] = typ

        if not isinstance(typ, TYPE_T):
            info = get_expr_info(node)  # get_expr_info fills in node._expr_info

            # log variable accesses.
            # (note writes will get logged as both read+write)
            var_access = _get_variable_access(node)
            if var_access is not None:
                info._reads.add(var_access)

            if self.function_analyzer:
                for s in self.function_analyzer.loop_variables:
                    if s is None:
                        continue

                    for v in info._writes:
                        if not v.contains(s):
                            continue

                        msg = "Cannot modify loop variable"
                        var = s.variable
                        if var.decl_node is not None:
                            msg += f" `{var.decl_node.target.id}`"
                        raise ImmutableViolation(msg, var.decl_node, node)

                var_accesses = info._writes | info._reads
                if uses_state(var_accesses):
                    self.function_analyzer._handle_module_access(node)

                self.func.mark_variable_writes(info._writes)
                self.func.mark_variable_reads(info._reads)

        # validate and annotate folded value
        if node.has_folded_value:
            folded_node = node.get_folded_value()
            self.visit(folded_node, typ)

    def visit_Attribute(self, node: vy_ast.Attribute, typ: VyperType) -> None:
        _validate_msg_data_attribute(node)

        # CMC 2023-10-19 TODO generalize this to mutability check on every node.
        # something like,
        # if self.func.mutability < expr_info.mutability:
        #    raise ...

        if self.func and self.func.mutability != StateMutability.PAYABLE:
            _validate_msg_value_access(node)

        if self.func and self.func.mutability == StateMutability.PURE:
            _validate_pure_access(node, typ)

        value_type = get_exact_type_from_node(node.value)

        _validate_address_code(node, value_type)

        self.visit(node.value, value_type)

    def visit_BinOp(self, node: vy_ast.BinOp, typ: VyperType) -> None:
        self.visit(node.left, typ)

        rtyp = typ
        if isinstance(node.op, (vy_ast.LShift, vy_ast.RShift)):
            rtyp = get_possible_types_from_node(node.right).pop()

        self.visit(node.right, rtyp)

    def visit_BoolOp(self, node: vy_ast.BoolOp, typ: VyperType) -> None:
        assert typ == BoolT()  # sanity check
        for value in node.values:
            self.visit(value, BoolT())

    def _check_call_mutability(self, call_mutability: StateMutability):
        # note: payable can be called from nonpayable functions
        ok = (
            call_mutability <= self.func.mutability
            or self.func.mutability >= StateMutability.NONPAYABLE
        )
        if not ok:
            msg = f"Cannot call a {call_mutability} function from a {self.func.mutability} function"
            raise StateAccessViolation(msg)

    def visit_ExtCall(self, node, typ):
        return self.visit(node.value, typ)

    def visit_StaticCall(self, node, typ):
        return self.visit(node.value, typ)

    def visit_Call(self, node: vy_ast.Call, typ: VyperType) -> None:
        func_info = get_expr_info(node.func, is_callable=True)
        func_type = func_info.typ

        # TODO: unify the APIs for different callable types so that
        # we don't need so much branching here.

        if not node.is_plain_call and not isinstance(func_type, ContractFunctionT):
            kind = node.kind_str
            msg = f"cannot use `{kind}` here!"
            hint = f"remove the `{kind}` keyword"
            raise CallViolation(msg, node.parent, hint=hint)

        if isinstance(func_type, ContractFunctionT):
            # function calls
            if func_type.is_external:
                missing_keyword = node.is_plain_call
                is_static = func_type.mutability < StateMutability.NONPAYABLE

                if is_static != node.is_staticcall or missing_keyword:
                    should = "staticcall" if is_static else "extcall"
                    msg = f"Calls to external {func_type.mutability} functions "
                    msg += f"must use the `{should}` keyword."
                    hint = f"try `{should} {node.node_source_code}`"
                    raise CallViolation(msg, hint=hint)
            else:
                if not node.is_plain_call:
                    kind = node.kind_str
                    msg = f"Calls to internal functions cannot use the `{kind}` keyword."
                    hint = f"remove the `{kind}` keyword"
                    raise CallViolation(msg, node.parent, hint=hint)

            if not func_type.from_interface:
                for s in func_type.get_variable_writes():
                    if s.variable.is_state_variable():
                        func_info._writes.add(s)
                for s in func_type.get_variable_reads():
                    if s.variable.is_state_variable():
                        func_info._reads.add(s)

            if self.function_analyzer:
                self._check_call_mutability(func_type.mutability)

                if func_type.uses_state():
                    self.function_analyzer._handle_module_access(node.func)

                if func_type.is_deploy and not self.func.is_deploy:
                    raise CallViolation(
                        f"Cannot call an @{func_type.visibility} function from "
                        f"an @{self.func.visibility} function!",
                        node,
                    )

            for arg, typ in zip(node.args, func_type.argument_types):
                self.visit(arg, typ)
            for kwarg in node.keywords:
                # We should only see special kwargs
                typ = func_type.call_site_kwargs[kwarg.arg].typ
                self.visit(kwarg.value, typ)

        elif is_type_t(func_type, EventT):
            # events have no kwargs
            expected_types = func_type.typedef.arguments.values()  # type: ignore
            for arg, typ in zip(node.args, expected_types):
                self.visit(arg, typ)
        elif is_type_t(func_type, StructT):
            # struct ctors
            # ctors have no kwargs
            expected_types = func_type.typedef.members.values()  # type: ignore
            for kwarg, arg_type in zip(node.keywords, expected_types):
                self.visit(kwarg.value, arg_type)
        elif isinstance(func_type, MemberFunctionT):
            if func_type.is_modifying and self.function_analyzer is not None:
                # TODO refactor this
                assert isinstance(node.func, vy_ast.Attribute)  # help mypy
                self.function_analyzer._handle_modification(node.func.value)
            assert len(node.args) == len(func_type.arg_types)
            for arg, arg_type in zip(node.args, func_type.arg_types):
                self.visit(arg, arg_type)
        else:
            # builtin functions
            arg_types = func_type.infer_arg_types(node, expected_return_typ=typ)  # type: ignore
            for arg, arg_type in zip(node.args, arg_types):
                self.visit(arg, arg_type)
            kwarg_types = func_type.infer_kwarg_types(node)  # type: ignore
            for kwarg in node.keywords:
                self.visit(kwarg.value, kwarg_types[kwarg.arg])

        self.visit(node.func, func_type)

    def visit_Compare(self, node: vy_ast.Compare, typ: VyperType) -> None:
        if isinstance(node.op, (vy_ast.In, vy_ast.NotIn)):
            # membership in list literal - `x in [a, b, c]`
            # needle: ltyp, haystack: rtyp
            if isinstance(node.right, vy_ast.List):
                ltyp = get_common_types(node.left, *node.right.elements).pop()

                rlen = len(node.right.elements)
                rtyp = SArrayT(ltyp, rlen)
            else:
                rtyp = get_exact_type_from_node(node.right)
                if isinstance(rtyp, FlagT):
                    # flag membership - `some_flag in other_flag`
                    ltyp = rtyp
                else:
                    # array membership - `x in my_list_variable`
                    assert isinstance(rtyp, (SArrayT, DArrayT))
                    ltyp = rtyp.value_type

            self.visit(node.left, ltyp)
            self.visit(node.right, rtyp)

        else:
            # ex. a < b
            cmp_typ = get_common_types(node.left, node.right).pop()
            if isinstance(cmp_typ, _BytestringT):
                # for bytestrings, get_common_types automatically downcasts
                # to the smaller common type - that will annotate with the
                # wrong type, instead use get_exact_type_from_node (which
                # resolves to the right type for bytestrings anyways).
                ltyp = get_exact_type_from_node(node.left)
                rtyp = get_exact_type_from_node(node.right)
            else:
                ltyp = rtyp = cmp_typ

            self.visit(node.left, ltyp)
            self.visit(node.right, rtyp)

    def visit_Constant(self, node: vy_ast.Constant, typ: VyperType) -> None:
        pass

    def visit_IfExp(self, node: vy_ast.IfExp, typ: VyperType) -> None:
        self.visit(node.test, BoolT())
        self.visit(node.body, typ)
        self.visit(node.orelse, typ)

    def visit_List(self, node: vy_ast.List, typ: VyperType) -> None:
        assert isinstance(typ, (SArrayT, DArrayT))
        for element in node.elements:
            self.visit(element, typ.value_type)

    def visit_Name(self, node: vy_ast.Name, typ: VyperType) -> None:
        if self.func and self.func.mutability == StateMutability.PURE:
            _validate_pure_access(node, typ)

    def visit_Subscript(self, node: vy_ast.Subscript, typ: VyperType) -> None:
        if isinstance(typ, TYPE_T):
            # don't recurse; can't annotate AST children of type definition
            return

        if isinstance(node.value, (vy_ast.List, vy_ast.Subscript)):
            possible_base_types = get_possible_types_from_node(node.value)

            for possible_type in possible_base_types:
                if typ.compare_type(possible_type.value_type):
                    base_type = possible_type
                    break
            else:
                # this should have been caught in
                # `get_possible_types_from_node` but wasn't.
                raise TypeCheckFailure(f"Expected {typ} but it is not a possible type", node)

        else:
            base_type = get_exact_type_from_node(node.value)

        if isinstance(base_type, HashMapT):
            index_type = base_type.key_type
        else:
            # Arrays allow most int types as index: Take the least specific
            index_type = get_possible_types_from_node(node.slice).pop()

        self.visit(node.value, base_type)
        self.visit(node.slice, index_type)

    def visit_Tuple(self, node: vy_ast.Tuple, typ: VyperType) -> None:
        if isinstance(typ, TYPE_T):
            # don't recurse; can't annotate AST children of type definition
            return

        # these guarantees should be provided by validate_expected_type
        assert isinstance(typ, TupleT)
        assert len(node.elements) == len(typ.member_types)

        for item_ast, item_type in zip(node.elements, typ.member_types):
            self.visit(item_ast, item_type)

    def visit_UnaryOp(self, node: vy_ast.UnaryOp, typ: VyperType) -> None:
        self.visit(node.operand, typ)


def _validate_range_call(node: vy_ast.Call):
    """
    Check that the arguments to a range() call are valid.
    :param node: call to range()
    :return: None
    """
    assert node.func.get("id") == "range"
    validate_call_args(node, (1, 2), kwargs=["bound"])
    kwargs = {s.arg: s.value for s in node.keywords or []}
    start, end = (vy_ast.Int(value=0), node.args[0]) if len(node.args) == 1 else node.args
    start, end = [i.reduced() for i in (start, end)]

    if "bound" in kwargs:
        bound = kwargs["bound"].reduced()
        if not isinstance(bound, vy_ast.Int):
            raise StructureException("Bound must be a literal integer", bound)
        if bound.value <= 0:
            raise StructureException("Bound must be at least 1", bound)
        if isinstance(start, vy_ast.Int) and isinstance(end, vy_ast.Int):
            error = "Please remove the `bound=` kwarg when using range with constants"
            raise StructureException(error, bound)
    else:
        for arg in (start, end):
            if not isinstance(arg, vy_ast.Int):
                error = "Value must be a literal integer, unless a bound is specified"
                raise StructureException(error, arg)
        if end.value <= start.value:
            raise StructureException("End must be greater than start", end)



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    """
@external
def foo():
    a: uint256 = 3
    b: int128 = 4
    c: uint256 = min(a, b)
    """,
    """
@external
def broken():
    a : uint256 = 3
    b : int128 = 4
    c : uint256 = unsafe_add(a, b)
    """,
    """
@external
def foo():
    b: Bytes[1] = b"\x05"
    x: uint256 = as_wei_value(b, "babbage")
    """,
    """
@external
def foo():
    raw_log(b"cow", b"dog")
    """,
    """
@external
def foo():
    xs: uint256[1] = []
    """,
    # literal longer than event member
    """
event Foo:
    message: String[1]
@external
def foo():
    log Foo("abcd")
    """,
    # Address literal must be checksummed
    """
a: constant(address) = 0x3cd751e6b0078be393132286c442345e5dc49699
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_type_mismatch_exception(bad_code):
    with raises(TypeMismatch):
        compiler.compile_code(bad_code)



import ast as python_ast

from vyper.ast.parse import annotate_python_ast, pre_parse


class AssertionVisitor(python_ast.NodeVisitor):
    def assert_about_node(self, node):
        raise AssertionError()

    def generic_visit(self, node):
        self.assert_about_node(node)

        super().generic_visit(node)


TEST_CONTRACT_SOURCE_CODE = """
struct S:
    a: bool
    b: int128

interface ERC20Contract:
    def name() -> String[64]: view

@external
def foo() -> int128:
    return -(-(-1))
"""


def get_contract_info(source_code):
    _, loop_var_annotations, class_types, reformatted_code = pre_parse(source_code)
    py_ast = python_ast.parse(reformatted_code)

    annotate_python_ast(py_ast, reformatted_code, loop_var_annotations, class_types)

    return py_ast, reformatted_code


def test_it_annotates_ast_with_source_code():
    contract_ast, reformatted_code = get_contract_info(TEST_CONTRACT_SOURCE_CODE)

    class AssertSourceCodePresent(AssertionVisitor):
        def assert_about_node(self, node):
            assert node.full_source_code is reformatted_code

    AssertSourceCodePresent().visit(contract_ast)


def test_it_annotates_ast_with_class_types():
    contract_ast, _ = get_contract_info(TEST_CONTRACT_SOURCE_CODE)

    struct_def = contract_ast.body[0]
    contract_def = contract_ast.body[1]

    assert struct_def.ast_type == "StructDef"
    assert contract_def.ast_type == "InterfaceDef"


def test_it_rewrites_unary_subtractions():
    contract_ast, _ = get_contract_info(TEST_CONTRACT_SOURCE_CODE)

    function_def = contract_ast.body[2]
    return_stmt = function_def.body[0]

    assert isinstance(return_stmt.value, python_ast.Constant)
    assert return_stmt.value.value == -1



import pytest

from tests.utils import parse_and_fold


@pytest.mark.parametrize("bool_cond", [True, False])
def test_unaryop(get_contract, bool_cond):
    source = """
@external
def foo(a: bool) -> bool:
    return not a
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"not {bool_cond}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(bool_cond) == new_node.value


@pytest.mark.parametrize("count", range(2, 11))
@pytest.mark.parametrize("bool_cond", [True, False])
def test_unaryop_nested(get_contract, bool_cond, count):
    source = f"""
@external
def foo(a: bool) -> bool:
    return {'not ' * count} a
    """
    contract = get_contract(source)

    literal_op = f"{'not ' * count}{bool_cond}"
    vyper_ast = parse_and_fold(literal_op)
    new_node = vyper_ast.body[0].value.get_folded_value()
    expected = new_node.value

    assert contract.foo(bool_cond) == expected



from typing import Tuple

from vyper.exceptions import StructureException, tag_exceptions


class VyperNodeVisitorBase:
    ignored_types: Tuple = ()
    scope_name = ""

    def visit(self, node, *args):
        if isinstance(node, self.ignored_types):
            return

        # iterate over the MRO until we find a matching visitor function
        # this lets us use a single function to broadly target several
        # node types with a shared parent
        for class_ in node.__class__.mro():
            ast_type = class_.__name__

            with tag_exceptions(node):
                visitor_fn = getattr(self, f"visit_{ast_type}", None)
                if visitor_fn:
                    return visitor_fn(node, *args)

        node_type = type(node).__name__
        raise StructureException(
            f"Unsupported syntax for {self.scope_name} namespace: {node_type}", node
        )



import pytest
from eth.codecs import abi

from tests.evm_backends.base_env import EvmError, ExecutionReverted
from tests.utils import decimal_to_int
from vyper.exceptions import ArgumentException, StructureException

TEST_ADDR = "0x" + b"".join(chr(i).encode("utf-8") for i in range(20)).hex()


def test_abi_decode_complex(get_contract):
    contract = """
struct Animal:
  name: String[5]
  address_: address
  id_: int128
  is_furry: bool
  price: decimal
  data: uint256[3]
  metadata: bytes32

struct Human:
  name: String[64]
  pet: Animal

@external
def abi_decode(x: Bytes[160]) -> (address, int128, bool, decimal, bytes32):
    a: address = empty(address)
    b: int128 = 0
    c: bool = False
    d: decimal = 0.0
    e: bytes32 = 0x0000000000000000000000000000000000000000000000000000000000000000
    a, b, c, d, e = _abi_decode(x, (address, int128, bool, decimal, bytes32))
    return a, b, c, d, e

@external
def abi_decode_struct(x: Bytes[544]) -> Human:
    human: Human = Human(
        name="",
        pet=Animal(
            name="",
            address_=empty(address),
            id_=0,
            is_furry=False,
            price=0.0,
            data=[0, 0, 0],
            metadata=0x0000000000000000000000000000000000000000000000000000000000000000
        )
    )
    human = _abi_decode(x, Human)
    return human
    """

    c = get_contract(contract)

    test_bytes32 = b"".join(chr(i).encode("utf-8") for i in range(32))
    args = (TEST_ADDR, -1, True, decimal_to_int("-123.4"), test_bytes32)
    encoding = "(address,int128,bool,int168,bytes32)"
    encoded = abi.encode(encoding, args)
    assert tuple(c.abi_decode(encoded)) == (
        TEST_ADDR,
        -1,
        True,
        decimal_to_int("-123.4"),
        test_bytes32,
    )

    test_bytes32 = b"".join(chr(i).encode("utf-8") for i in range(32))
    human_tuple = (
        "foobar",
        ("vyper", TEST_ADDR, 123, True, decimal_to_int("123.4"), [123, 456, 789], test_bytes32),
    )

    human_t = "((string,(string,address,int128,bool,int168,uint256[3],bytes32)))"
    human_encoded = abi.encode(human_t, (human_tuple,))
    assert c.abi_decode_struct(human_encoded) == (
        "foobar",
        ("vyper", TEST_ADDR, 123, True, decimal_to_int("123.4"), [123, 456, 789], test_bytes32),
    )


@pytest.mark.parametrize(
    "expected,input_len,output_typ,abi_typ,unwrap_tuple",
    [
        (123, 32, "uint256", "uint256", False),
        (123, 32, "uint256", "(uint256)", True),
        ("vyper", 64, "String[5]", "string", False),
        ("vyper", 96, "String[5]", "(string)", True),
        ([123, 456, 789], 96, "uint256[3]", "uint256[3]", False),
        ([123, 456, 789], 96, "uint256[3]", "(uint256[3])", True),
        ([123, 456, 789], 128, "DynArray[uint256, 3]", "uint256[]", False),
        ([123, 456, 789], 160, "DynArray[uint256, 3]", "(uint256[])", True),
    ],
)
def test_abi_decode_single(get_contract, expected, input_len, output_typ, abi_typ, unwrap_tuple):
    contract = f"""
@external
def foo(x: Bytes[{input_len}]) -> {output_typ}:
    a: {output_typ} = _abi_decode(x, {output_typ}, unwrap_tuple={unwrap_tuple})
    return a
    """
    c = get_contract(contract)

    encode_arg = expected
    if unwrap_tuple is True:
        encode_arg = (expected,)

    encoded = abi.encode(abi_typ, encode_arg)
    assert c.foo(encoded) == expected


@pytest.mark.parametrize(
    "arg,expected,input_len,output_typ1,output_typ2,abi_typ",
    [
        ((123, 456), (123, 456), 64, "uint256", "uint256", "(uint256,uint256)"),
        ((TEST_ADDR, 123), (TEST_ADDR, 123), 64, "address", "int128", "(address,int128)"),
        (
            ("vyper", TEST_ADDR),
            ("vyper", TEST_ADDR),
            128,
            "String[5]",
            "address",
            "(string,address)",
        ),
        ((1, b"234"), (1, b"234"), 128, "uint256", "Bytes[32]", ("(uint256,bytes)")),
    ],
)
@pytest.mark.parametrize("unwrap_tuple", (True, False))
def test_abi_decode_double(
    get_contract, arg, expected, input_len, output_typ1, output_typ2, abi_typ, unwrap_tuple
):
    contract = f"""
@external
def foo(x: Bytes[{input_len}]) -> ({output_typ1}, {output_typ2}):
    a: {output_typ1} = empty({output_typ1})
    b: {output_typ2} = empty({output_typ2})
    a, b = _abi_decode(x, ({output_typ1}, {output_typ2}), unwrap_tuple={unwrap_tuple})
    return a, b
    """

    c = get_contract(contract)
    encoded = abi.encode(abi_typ, arg)
    assert tuple(c.foo(encoded)) == expected


nested_2d_array_args = [
    [[123, 456, 789], [234, 567, 891], [345, 678, 912]],
    [[], [], []],
    [[123, 456], [234, 567, 891]],
    [[123, 456, 789], [234, 567], [345]],
    [[123], [], [345, 678, 912]],
    [[], [], [345, 678, 912]],
    [[], [], [345]],
    [[], [234], []],
    [[], [234, 567, 891], []],
    [[]],
    [[123], [234]],
]


@pytest.mark.parametrize("args", nested_2d_array_args)
@pytest.mark.parametrize("unwrap_tuple", (True, False))
def test_abi_decode_nested_dynarray(get_contract, args, unwrap_tuple):
    if unwrap_tuple is True:
        encoded = abi.encode("(uint256[][])", (args,))
        len = 544
    else:
        encoded = abi.encode("uint256[][]", args)
        len = 512

    code = f"""
@external
def abi_decode(x: Bytes[{len}]) -> DynArray[DynArray[uint256, 3], 3]:
    a: DynArray[DynArray[uint256, 3], 3] = []
    a = _abi_decode(x, DynArray[DynArray[uint256, 3], 3], unwrap_tuple={unwrap_tuple})
    return a
    """

    c = get_contract(code)
    assert c.abi_decode(encoded) == args


nested_3d_array_args = [
    [
        [[123, 456, 789], [234, 567, 891], [345, 678, 912]],
        [[234, 567, 891], [345, 678, 912], [123, 456, 789]],
        [[345, 678, 912], [123, 456, 789], [234, 567, 891]],
    ],
    [[[123, 789], [234], [345, 678, 912]], [[234, 567], [345, 678]], [[345]]],
    [[[123], [234, 567, 891]], [[234]]],
    [[[], [], []], [[], [], []], [[], [], []]],
    [[[123, 456, 789], [234, 567, 891], [345, 678, 912]], [[234, 567, 891], [345, 678, 912]], [[]]],
    [[[]], [[]], [[234]]],
    [[[123]], [[]], [[]]],
    [[[]], [[123]], [[]]],
    [[[123, 456, 789], [234, 567]], [[234]], [[567], [912], [345]]],
    [[[]]],
]


@pytest.mark.parametrize("args", nested_3d_array_args)
@pytest.mark.parametrize("unwrap_tuple", (True, False))
def test_abi_decode_nested_dynarray2(get_contract, args, unwrap_tuple):
    if unwrap_tuple is True:
        encoded = abi.encode("(uint256[][][])", (args,))
        len = 1696
    else:
        encoded = abi.encode("uint256[][][]", args)
        len = 1664

    code = f"""
@external
def abi_decode(x: Bytes[{len}]) -> DynArray[DynArray[DynArray[uint256, 3], 3], 3]:
    a: DynArray[DynArray[DynArray[uint256, 3], 3], 3] = []
    a = _abi_decode(
        x,
        DynArray[DynArray[DynArray[uint256, 3], 3], 3],
        unwrap_tuple={unwrap_tuple}
    )
    return a
    """

    c = get_contract(code)
    assert c.abi_decode(encoded) == args


def test_side_effects_evaluation(get_contract):
    contract_1 = """
counter: uint256

@deploy
def __init__():
    self.counter = 0

@external
def get_counter() -> Bytes[128]:
    self.counter += 1
    return _abi_encode(self.counter, "hello")
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def get_counter() -> Bytes[128]: nonpayable

@external
def foo(addr: address) -> (uint256, String[5]):
    a: uint256 = 0
    b: String[5] = ""
    a, b = _abi_decode(extcall Foo(addr).get_counter(), (uint256, String[5]), unwrap_tuple=False)
    return a, b
    """

    c2 = get_contract(contract_2)

    assert tuple(c2.foo(c.address)) == (1, "hello")


def test_abi_decode_private_dynarray(get_contract):
    code = """
bytez: DynArray[uint256, 3]

@internal
def _foo(bs: Bytes[160]):
    self.bytez = _abi_decode(bs, DynArray[uint256, 3])

@external
def foo(bs: Bytes[160]) -> (uint256, DynArray[uint256, 3]):
    dont_clobber_me: uint256 = max_value(uint256)
    self._foo(bs)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    bs = [1, 2, 3]
    encoded = abi.encode("(uint256[])", (bs,))
    assert c.foo(encoded) == (2**256 - 1, bs)


def test_abi_decode_private_nested_dynarray(get_contract):
    code = """
bytez: DynArray[DynArray[DynArray[uint256, 3], 3], 3]

@internal
def _foo(bs: Bytes[1696]):
    self.bytez = _abi_decode(bs, DynArray[DynArray[DynArray[uint256, 3], 3], 3])

@external
def foo(bs: Bytes[1696]) -> (uint256, DynArray[DynArray[DynArray[uint256, 3], 3], 3]):
    dont_clobber_me: uint256 = max_value(uint256)
    self._foo(bs)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    bs = [
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        [[10, 11, 12], [13, 14, 15], [16, 17, 18]],
        [[19, 20, 21], [22, 23, 24], [25, 26, 27]],
    ]
    encoded = abi.encode("(uint256[][][])", (bs,))
    assert c.foo(encoded) == (2**256 - 1, bs)


def test_abi_decode_return(get_contract):
    contract = """
@external
def abi_decode(x: Bytes[64]) -> (address, int128):
    return _abi_decode(x, (address, int128))
    """

    c = get_contract(contract)

    encoded = abi.encode("(address,int128)", (TEST_ADDR, 123))

    assert tuple(c.abi_decode(encoded)) == (TEST_ADDR, 123)


def test_abi_decode_annassign(get_contract):
    contract = """
@external
def abi_decode(x: Bytes[32]) -> uint256:
    a: uint256 = _abi_decode(x, uint256, unwrap_tuple=False)
    return a
    """
    c = get_contract(contract)

    encoded = abi.encode("uint256", 123)
    assert c.abi_decode(encoded) == 123


@pytest.mark.parametrize(
    "input_",
    [
        b"",  # Length of byte array is below minimum size of output type
        b"\x01" * 96,  # Length of byte array is beyond size bound of output type
    ],
)
def test_clamper(get_contract, tx_failed, input_):
    contract = """
@external
def abi_decode(x: Bytes[96]) -> (uint256, uint256):
    a: uint256 = empty(uint256)
    b: uint256 = empty(uint256)
    a, b = _abi_decode(x, (uint256, uint256))
    return a, b
    """
    c = get_contract(contract)
    with tx_failed():
        c.abi_decode(input_)


def test_clamper_nested_uint8(get_contract, tx_failed):
    # check that _abi_decode clamps on word-types even when it is in a nested expression
    # decode -> validate uint8 -> revert if input >= 256 -> cast back to uint256
    contract = """
@external
def abi_decode(x: uint256) -> uint256:
    a: uint256 = convert(_abi_decode(slice(msg.data, 4, 32), (uint8)), uint256)
    return a
    """
    c = get_contract(contract)
    assert c.abi_decode(255) == 255
    with tx_failed():
        c.abi_decode(256)


def test_clamper_nested_bytes(get_contract, tx_failed):
    # check that _abi_decode clamps dynamic even when it is in a nested expression
    # decode -> validate Bytes[20] -> revert if len(input) > 20 -> convert back to -> add 1
    contract = """
@external
def abi_decode(x: Bytes[96]) -> Bytes[21]:
    a: Bytes[21] = concat(b"a", _abi_decode(x, Bytes[20]))
    return a
    """
    c = get_contract(contract)
    assert c.abi_decode(abi.encode("(bytes)", (b"bc",))) == b"abc"
    with tx_failed():
        c.abi_decode(abi.encode("(bytes)", (b"a" * 22,)))


@pytest.mark.parametrize(
    "output_typ,input_",
    [
        ("DynArray[uint256, 3]", b""),
        ("DynArray[uint256, 3]", b"\x01" * 192),
        ("Bytes[5]", b""),
        ("Bytes[5]", b"\x01" * 192),
    ],
)
def test_clamper_dynamic(get_contract, tx_failed, output_typ, input_):
    contract = f"""
@external
def abi_decode(x: Bytes[192]) -> {output_typ}:
    a: {output_typ} = empty({output_typ})
    a = _abi_decode(x, {output_typ})
    return a
    """
    c = get_contract(contract)
    with tx_failed():
        c.abi_decode(input_)


@pytest.mark.parametrize(
    "arg,encoding,expected", [(123, "(uint256)", 123), ([123, 456, 789], "(uint256[])", 789)]
)
def test_abi_decode_conditional(get_contract, arg, encoding, expected):
    contract = """
@external
def abi_decode(x: Bytes[160]) -> uint256:
    if len(x) == 32:
        a: uint256 = _abi_decode(x, uint256)
        return a
    elif len(x) == 160:
        b: DynArray[uint256, 3] = _abi_decode(x, DynArray[uint256, 3])
        return b[2]
    return 0
    """
    c = get_contract(contract)
    encoded = abi.encode(encoding, (arg,))
    assert c.abi_decode(encoded) == expected


@pytest.mark.parametrize(
    "output_typ1,output_typ2,input_,error,error_property",
    [
        ("DynArray[uint256, 3]", "uint256", b"", ExecutionReverted, ""),
        ("DynArray[uint256, 3]", "uint256", b"\x01" * 128, EvmError, "OUT_OF_GAS_ERROR"),
        ("Bytes[5]", "address", b"", ExecutionReverted, ""),
        ("Bytes[5]", "address", b"\x01" * 128, EvmError, "OUT_OF_GAS_ERROR"),
    ],
)
def test_clamper_dynamic_tuple(
    get_contract, tx_failed, output_typ1, output_typ2, input_, error, error_property, env
):
    contract = f"""
@external
def abi_decode(x: Bytes[224]) -> ({output_typ1}, {output_typ2}):
    a: {output_typ1} = empty({output_typ1})
    b: {output_typ2} = empty({output_typ2})
    a, b = _abi_decode(x, ({output_typ1}, {output_typ2}))
    return a, b
    """
    c = get_contract(contract)
    with tx_failed(error, exc_text=getattr(env, error_property, None)):
        c.abi_decode(input_)


FAIL_LIST = [
    (
        """
@external
def foo(x: Bytes[32]):
    a: uint256 = 0
    b: uint256 = 0
    a, b = _abi_decode(x, (uint256, uint256))
    """,
        StructureException,  # Size of input data is smaller than expected output
    ),
    (
        """
@external
def foo(x: Bytes[32]):
    _abi_decode(x)
    """,
        ArgumentException,  # Output types arg is missing
    ),
]


@pytest.mark.parametrize("bad_code,exception", FAIL_LIST)
def test_abi_decode_length_mismatch(get_contract, assert_compile_failed, bad_code, exception):
    assert_compile_failed(lambda: get_contract(bad_code), exception)



import cbor2
import pytest

import vyper
from vyper.compiler.settings import OptimizationLevel, Settings

simple_contract_code = """
@external
def a() -> bool:
    return True
"""

many_functions = """
@external
def foo1():
    pass

@external
def foo2():
    pass

@external
def foo3():
    pass

@external
def foo4():
    pass

@external
def foo5():
    pass
"""

has_immutables = """
A_GOOD_PRIME: public(immutable(uint256))

@deploy
def __init__():
    A_GOOD_PRIME = 967
"""


def _parse_cbor_metadata(initcode):
    metadata_ofst = int.from_bytes(initcode[-2:], "big")
    metadata = cbor2.loads(initcode[-metadata_ofst:-2])
    return metadata


def test_bytecode_runtime():
    out = vyper.compile_code(simple_contract_code, output_formats=["bytecode_runtime", "bytecode"])

    assert len(out["bytecode"]) > len(out["bytecode_runtime"])
    assert out["bytecode_runtime"].removeprefix("0x") in out["bytecode"].removeprefix("0x")


def test_bytecode_signature():
    out = vyper.compile_code(simple_contract_code, output_formats=["bytecode_runtime", "bytecode"])

    runtime_code = bytes.fromhex(out["bytecode_runtime"].removeprefix("0x"))
    initcode = bytes.fromhex(out["bytecode"].removeprefix("0x"))

    metadata = _parse_cbor_metadata(initcode)
    runtime_len, data_section_lengths, immutables_len, compiler = metadata

    assert runtime_len == len(runtime_code)
    assert data_section_lengths == []
    assert immutables_len == 0
    assert compiler == {"vyper": list(vyper.version.version_tuple)}


def test_bytecode_signature_dense_jumptable():
    settings = Settings(optimize=OptimizationLevel.CODESIZE)

    out = vyper.compile_code(
        many_functions, output_formats=["bytecode_runtime", "bytecode"], settings=settings
    )

    runtime_code = bytes.fromhex(out["bytecode_runtime"].removeprefix("0x"))
    initcode = bytes.fromhex(out["bytecode"].removeprefix("0x"))

    metadata = _parse_cbor_metadata(initcode)
    runtime_len, data_section_lengths, immutables_len, compiler = metadata

    assert runtime_len == len(runtime_code)
    assert data_section_lengths == [5, 35]
    assert immutables_len == 0
    assert compiler == {"vyper": list(vyper.version.version_tuple)}


def test_bytecode_signature_sparse_jumptable():
    settings = Settings(optimize=OptimizationLevel.GAS)

    out = vyper.compile_code(
        many_functions, output_formats=["bytecode_runtime", "bytecode"], settings=settings
    )

    runtime_code = bytes.fromhex(out["bytecode_runtime"].removeprefix("0x"))
    initcode = bytes.fromhex(out["bytecode"].removeprefix("0x"))

    metadata = _parse_cbor_metadata(initcode)
    runtime_len, data_section_lengths, immutables_len, compiler = metadata

    assert runtime_len == len(runtime_code)
    assert data_section_lengths == [8]
    assert immutables_len == 0
    assert compiler == {"vyper": list(vyper.version.version_tuple)}


def test_bytecode_signature_immutables():
    out = vyper.compile_code(has_immutables, output_formats=["bytecode_runtime", "bytecode"])

    runtime_code = bytes.fromhex(out["bytecode_runtime"].removeprefix("0x"))
    initcode = bytes.fromhex(out["bytecode"].removeprefix("0x"))

    metadata = _parse_cbor_metadata(initcode)
    runtime_len, data_section_lengths, immutables_len, compiler = metadata

    assert runtime_len == len(runtime_code)
    assert data_section_lengths == []
    assert immutables_len == 32
    assert compiler == {"vyper": list(vyper.version.version_tuple)}


# check that deployed bytecode actually matches the cbor metadata
@pytest.mark.parametrize("code", [simple_contract_code, has_immutables, many_functions])
def test_bytecode_signature_deployed(code, get_contract, env):
    c = get_contract(code)
    deployed_code = env.get_code(c.address)

    metadata = _parse_cbor_metadata(c.bytecode)
    runtime_len, data_section_lengths, immutables_len, compiler = metadata

    assert compiler == {"vyper": list(vyper.version.version_tuple)}

    # runtime_len includes data sections but not immutables
    assert len(deployed_code) == runtime_len + immutables_len



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import FunctionDeclarationException

good_list = [
    """
@deploy
def __init__():
    pass
    """,
    """
@deploy
@payable
def __init__():
    pass
    """,
    """
counter: uint256
SOME_IMMUTABLE: immutable(uint256)

@deploy
def __init__():
    SOME_IMMUTABLE = 5
    self.counter = 1
    """,
]


@pytest.mark.parametrize("code", good_list)
def test_good_init_funcs(code):
    assert compile_code(code) is not None


fail_list = [
    """
@internal
def __init__():
    pass
    """,
    """
@deploy
@view
def __init__():
    pass
    """,
    """
@deploy
@pure
def __init__():
    pass
    """,
    """
@deploy
def some_function():  # for now, only __init__() functions can be marked @deploy
    pass
    """,
]


@pytest.mark.parametrize("code", fail_list)
def test_bad_init_funcs(code):
    with pytest.raises(FunctionDeclarationException):
        compile_code(code)



import os

import pytest

from vyper.cli.vyper_compile import _parse_args


@pytest.fixture
def chdir_path(tmp_path):
    orig_path = os.getcwd()
    yield tmp_path
    os.chdir(orig_path)


def test_paths(chdir_path):
    code = """
@external
def foo() -> bool:
    return True
"""
    bar_path = chdir_path.joinpath("bar.vy")
    with bar_path.open("w") as fp:
        fp.write(code)

    _parse_args([str(bar_path)])  # absolute path
    os.chdir(chdir_path.parent)

    _parse_args([str(bar_path)])  # absolute path, subfolder of cwd
    _parse_args([str(bar_path.relative_to(chdir_path.parent))])  # relative path
